from .activity_details import ActivityDetailsPage
from .welcome import WelcomePage
from .calculation_setup import CalculationSetupPage
from .impact_category_details import ImpactCategoryDetailsPage
from .lca_results import LCAResultsPage
from .parameters import ParametersPage
from .metadatastore import MetaDataStorePage
from .settings import SettingsPage

base_pages = {
    "Welcome": WelcomePage,
    "Parameters": ParametersPage,
    "Settings": SettingsPage,
}



# -*- coding: utf-8 -*-
__all__ = ["panes", "pages", "application", "signals", "metadata", "main_window", "actions"]

import os

from activity_browser.ui.core.application import ABApplication
from activity_browser.bwutils.metadata import MetaDataStore
from activity_browser.bwutils.settings import Settings
from .main import MainWindow

application = ABApplication()
metadata = MetaDataStore(application)
settings = Settings()

# modules dependent on application instance
from .signalling import ABSignals

signals = ABSignals()

# modules dependent on application and signals
from . import actions
from . import panes
from . import pages
from . import dialogs

main_window = MainWindow()
application.main_window = main_window

if not os.environ.get("AB_SKIP_SETTINGS_ON_STARTUP"):
    main_window.apply_settings(load=True)  # Ensure settings are applied at startup




from PySide6.QtCore import QModelIndex
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt

from loguru import logger

import pandas as pd

from bw2io.importers.base_lci import LCIImporter

from activity_browser.ui import widgets, core, delegates, icons

from ..node_select_dialog import NodeSelectDialog


class ImportPreviewEdgeTab(QtWidgets.QWidget):
    standardEdgeColumns = ["linked", "type", "amount", "unit", "input", "name", "location", "database", "formula"]

    def __init__(self, importer: LCIImporter, parent=None):
        super().__init__(parent)
        self.importer = importer
        self.simple = True
        self.old_links: dict[tuple[int, int], tuple[str, str] | None] = {}

        layout = QtWidgets.QVBoxLayout(self)

        self.edge_model = ImportPreviewEdgeModel(parent=self)
        self.edge_model.set_dataframe(self.build_df())
        self.edge_model.group(["_node"])

        self.edge_view = ImportPreviewEdgeView(importer, self)
        self.edge_view.setUniformRowHeights(False)
        self.edge_view.setModel(self.edge_model)
        self.edge_view.setColumnWidth(0, 0)

        # Create simple/detailed view toggle
        self.view_toggle = QtWidgets.QCheckBox("Details")
        self.view_toggle.setChecked(not self.simple)
        self.view_toggle.setToolTip("Toggle between simple and detailed view")
        self.view_toggle.checkStateChanged.connect(self.on_mode_switch)

        # Create top bar with toggle
        top_bar = QtWidgets.QHBoxLayout()
        top_bar.addStretch()
        top_bar.addWidget(self.view_toggle)

        layout.addLayout(top_bar)
        layout.addWidget(self.edge_view)

        self.sync()

    def sync(self):
        """Synchronize the view based on simple/detailed mode."""
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        self.edge_view.header().setHidden(self.simple)
        self.edge_view.viewport().setBackgroundRole(
            QtGui.QPalette.ColorRole.Window if self.simple else QtGui.QPalette.ColorRole.Base)
        self.edge_view.setFrameShape(
            QtWidgets.QFrame.Shape.NoFrame if self.simple else QtWidgets.QFrame.Shape.StyledPanel)

        df = self.build_df()

        if self.simple and "_exc" in df.columns:
            df.rename(columns={"_exc": "exc"}, inplace=True)
        elif not self.simple and "node" in df.columns:
            df.rename(columns={"exc": "_exc"}, inplace=True)

        self.edge_model.update_dataframe(df)

        for col in self.edge_model.columns():
            if col == "index":
                continue
            index = self.edge_model.columns().index(col)

            hidden = (self.simple and not col == "exc") or (not self.simple and col == "exc")
            self.edge_view.setColumnHidden(index, hidden)

    def build_df(self):

        exchanges = []
        for node_i, node in enumerate(self.importer.data):
            summary = [
                    node.get("name"),
                    node.get("location"),
                    node.get("database"),
                    node.get("code"),
            ]
            summary = " | ".join([str(part) for part in summary if part])

            for exc_i, exc in enumerate(node.get("exchanges", [])):
                exc = exc.copy()
                exc["_node"] = summary
                exc["_location"] = (node_i, exc_i)
                exchanges.append(exc)

        df = pd.DataFrame(exchanges)
        for col in [col for col in self.standardEdgeColumns if col not in df.columns]:
            df[col] = None
        df["exc"] = None

        def determine_link_status(row):
            input_val = row["input"]
            location = row["_location"]

            if not isinstance(input_val, tuple):
                return "unlinked"
            elif location in self.old_links:
                return "relinked"
            else:
                return "linked"

        df["linked"] = df.apply(determine_link_status, axis=1)

        return df

    def on_mode_switch(self, check: Qt.CheckState):
        """Handle the mode switch between simple and detailed view."""
        self.simple = check == Qt.CheckState.Unchecked
        self.sync()

    def relink_selected_exchanges(self):
        """Open a dialog to link selected exchanges to existing nodes."""
        exchange_locations = self.edge_view.selected_exchanges
        if not exchange_locations:
            return

        dialog = NodeSelectDialog(parent=self)
        if not dialog.exec() == QtWidgets.QDialog.DialogCode.Accepted:
            return

        selected_node = dialog.get_selected_node()

        for loc in exchange_locations:
            node_i, exc_i = loc

            if loc not in self.old_links:
                self.old_links[loc] = self.importer.data[node_i]["exchanges"][exc_i].get("input")

            self.importer.data[node_i]["exchanges"][exc_i]["input"] = (selected_node["database"], selected_node["code"])

        self.sync()


class ShiftedCardDelegate(delegates.CardDelegate):
    """
    Delegate that shifts the card content to the left to compensate for indentation.
    """
    def paint(self, painter, option, index):
        # Adjust the rect to shift content left, compensating for indentation
        adjusted_option = QtWidgets.QStyleOptionViewItem(option)
        adjusted_option.rect.adjust(-28, 0, 0, 0)

        # Call the original paint with adjusted rect
        super().paint(painter, adjusted_option, index)


class ImportPreviewEdgeView(widgets.ABTreeView):
    """View for displaying import preview nodes."""

    defaultColumnDelegates = {
        "exc": ShiftedCardDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.callback(
                text="Link exchange" if len(p.selected_exchanges) == 1 else "Link exchanges",
                func=p.tab.relink_selected_exchanges,
            )
        ]

    def __init__(self, importer: LCIImporter, tab: ImportPreviewEdgeTab):
        super().__init__(tab)
        self.importer = importer
        self.old_links = {}
        self.tab = tab

    @property
    def selected_exchanges(self):
        """
        Returns a list of selected exchange locations as (node_index, exchange_index) tuples. These can be used to
        identify and manipulate the selected exchanges in the importer's data, which is a list of lists.
        """
        return list(set([self.model().get(index, "_location") for index in self.selectedIndexes()]))


class ImportPreviewEdgeModel(core.ABTreeModel):
    """Model for import preview nodes with node delegate support."""

    def displayData(self, index: QtCore.QModelIndex) -> any:
        if not index.isValid():
            return None

        column_name = self.columns()[index.column()]
        if not column_name == "exc" or self.row(index) is None:
            return super().displayData(index)

        row_data = self.row(index).copy()
        row_data.dropna(inplace=True)

        # Build the card information
        title = row_data.get('reference product') or row_data.get('name')
        subtitle = row_data.get('name')
        detail = f"{row_data.get('amount')} {row_data.get('unit')}"

        # Build categories list from unit, location
        categories = []
        if row_data.get("type"):
            categories.append(str(row_data.get("type")))
        if row_data.get("location"):
            categories.append(str(row_data.get("location")))
        if row_data.get("categories"):
            categories.append(", ".join([str(cat) for cat in row_data.get("categories")]))
        if row_data.get("database"):
            categories.append(str(row_data.get("database")))

        return {
            "title": title,
            "subtitle": subtitle,
            "categories": categories if categories else None,
            "detail": detail,
        }


    def decorationData(self, index: QModelIndex) -> QtGui.QIcon:
        if not index.isValid():
            return icons.qicons.empty

        column_name = self.columns()[index.column()]
        if not column_name in ["exc"]:
            return super().decorationData(index)

        linked = self.get(index, "linked")
        if linked == "linked":
            return icons.qicons.link
        elif linked == "unlinked":
            return icons.qicons.unlink
        elif linked == "relinked":
            return icons.qicons.relink
        return icons.qicons.empty

    def indexSelectable(self, index: QModelIndex) -> bool:
        # Don't make the tree column selectable
        if index.column() == 0:
            return False
        return True









from uuid import uuid4

from qtpy import QtWidgets
import bw2data as bd

from activity_browser import app
from activity_browser.bwutils.commontasks import database_is_legacy
from activity_browser.app.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons

from .activity_open import ActivityOpen


class ActivityNewProcess(ABAction):
    """
    ABAction to create a new activity. Prompts the user to supply a name. Returns if no name is supplied or if the user
    cancels. Otherwise, instructs the ActivityController to create a new activity.
    """

    icon = qicons.add
    text = "New process"

    @staticmethod
    @exception_dialogs
    def run(database_name: str):
        # ask the user to provide a name for the new activity
        dialog = NewNodeDialog(app.main_window)
        # if the user cancels, return
        if dialog.exec_() != QtWidgets.QDialog.DialogCode.Accepted:
            return
        name, ref_product, unit, location = dialog.get_new_process_data()
        # if no name is provided, return
        if not name:
            return
        if ref_product == "":
            ref_product = name

        database = bd.Database(database_name)
        legacy_backend = database_is_legacy(database_name)

        # create process
        new_proc_data = {
            "name": name,
            "location": location,
            "type": "process" if not legacy_backend else "processwithreferenceproduct",
        }

        if legacy_backend:
            new_proc_data["reference product"] = ref_product
            new_proc_data["unit"] = unit

        new_process: bd.Node = database.new_activity(code=uuid4().hex, **new_proc_data)
        new_process.save()

        if legacy_backend:
            new_process.new_exchange(
                input=new_process.key,
                type="production",
                amount=1.0,
            ).save()

        if not legacy_backend:
            # create reference product
            new_ref_prod_data = {
                "product": ref_product,
                "unit": unit,
                "location": location,
                "type": "product",
            }
            prod = new_process.new_product(code=uuid4().hex, **new_ref_prod_data)
            prod.save()

        ActivityOpen.run([new_process.key])


class NewNodeDialog(QtWidgets.QDialog):
    """
    Gathers the paremeters for creating a new process.
    """

    def __init__(self, process: bool = True, parent = None):
        super().__init__(parent)
        layout = QtWidgets.QGridLayout()
        row = 0
        if process:
            self.setWindowTitle("New process")
            layout.addWidget(QtWidgets.QLabel("Process name"), row, 0)
        else:
            self.setWindowTitle("New product")
            layout.addWidget(QtWidgets.QLabel("Product name"), row, 0)
        self._process_name_edit = QtWidgets.QLineEdit()
        self._process_name_edit.textChanged.connect(self._handle_text_changed)
        layout.addWidget(self._process_name_edit, row, 1)
        row += 1
        self._ref_product_name_edit = QtWidgets.QLineEdit()
        if process:
            layout.addWidget(QtWidgets.QLabel("Product name"), row, 0)
            layout.addWidget(self._ref_product_name_edit, row, 1)
            row += 1
        layout.addWidget(QtWidgets.QLabel("Unit"), row, 0)
        self._unit_edit = QtWidgets.QLineEdit("kilogram")
        layout.addWidget(self._unit_edit, row, 1)
        row += 1
        layout.addWidget(QtWidgets.QLabel("Location"), row, 0)
        default_loc = "GLO" if process else ""
        self._location_edit = QtWidgets.QLineEdit(default_loc)
        layout.addWidget(self._location_edit, row, 1)
        row += 1
        self._ok_button = QtWidgets.QPushButton("OK")
        self._ok_button.clicked.connect(self.accept)
        self._ok_button.setEnabled(False)
        layout.addWidget(self._ok_button, row, 0)
        cancel_button = QtWidgets.QPushButton("Cancel")
        cancel_button.clicked.connect(self.reject)
        layout.addWidget(cancel_button, row, 1)
        self.setLayout(layout)

    def _handle_text_changed(self, text: str):
        self._ok_button.setEnabled(text != "")
        self._ref_product_name_edit.setPlaceholderText(text)

    def get_new_process_data(self) -> tuple[str, str, str, str]:
        """Return the parameters the user entered."""
        return (
                self._process_name_edit.text(),
                self._ref_product_name_edit.text(),
                self._unit_edit.text(),
                self._location_edit.text()
            )




# -*- coding: utf-8 -*-
from loguru import logger
from qtpy import QtWidgets

from activity_browser.app import settings
from activity_browser.app.actions.metadatastore_cache_clear import MetaDataStoreCacheClear
from activity_browser.app.pages.settings.base import BaseSettingsChapter


class MetadataStoreSettingsChapter(BaseSettingsChapter):
    """Chapter for metadatastore-related settings."""

    def __init__(self, parent=None):
        super().__init__(parent)

        # Caching enabled checkbox
        self.caching_checkbox = QtWidgets.QCheckBox("Enable caching")
        self.caching_checkbox.setToolTip(
            "Enable caching for faster data access. "
            "Disable if you experience memory issues or want to force fresh data loading."
        )

        # Searcher enabled checkbox
        self.searcher_checkbox = QtWidgets.QCheckBox("Enable searcher")
        self.searcher_checkbox.setToolTip(
            "Enable the full-text search functionality for activities and metadata. "
            "Disable if you experience performance issues with large databases."
        )

        # Clear cache button
        self.clear_cache_button = QtWidgets.QPushButton("Clear Cache")
        self.clear_cache_button.setToolTip(
            "Clear the metadata store cache and reload the current project. "
            "Use this if you experience issues with outdated or corrupted cache data."
        )

        self.build_layout()
        self.connect_signals()
        self.reset()

    def connect_signals(self):
        """Connect signals and slots."""
        # Emit changed signal when settings change
        self.caching_checkbox.stateChanged.connect(lambda: self.changed.emit())
        self.searcher_checkbox.stateChanged.connect(lambda: self.changed.emit())

        # Connect clear cache button
        self.clear_cache_button.clicked.connect(MetaDataStoreCacheClear.run)

    def build_layout(self):
        """Build the chapter layout."""
        layout = QtWidgets.QVBoxLayout()

        # Metadata store group
        metadatastore_group = QtWidgets.QGroupBox("Metadata Store Options")
        metadatastore_layout = QtWidgets.QVBoxLayout()

        metadatastore_layout.addWidget(self.caching_checkbox)
        metadatastore_layout.addWidget(self.searcher_checkbox)

        # Add clear cache button
        metadatastore_layout.addWidget(self.clear_cache_button)

        # Add description label
        description = QtWidgets.QLabel(
            "These settings control the behavior of the metadata store, "
            "which manages activity data for improved performance."
        )
        description.setWordWrap(True)
        description.setStyleSheet("color: gray; font-size: 10pt;")
        metadatastore_layout.addWidget(description)

        metadatastore_group.setLayout(metadatastore_layout)

        layout.addWidget(metadatastore_group)
        layout.addStretch()

        self.setLayout(layout)

    # --- Settings management methods --- #
    def reset(self):
        """(Re)set to initial values."""
        try:
            self.caching_checkbox.setChecked(
                settings["metadatastore"]["caching_enabled"]
            )
            self.searcher_checkbox.setChecked(
                settings["metadatastore"]["searcher_enabled"]
            )
        except (KeyError, TypeError):
            # Use defaults if settings don't exist yet
            self.caching_checkbox.setChecked(True)
            self.searcher_checkbox.setChecked(True)

    def has_changes(self):
        """Check if there are unsaved changes."""
        try:
            current_state = {
                'caching_enabled': self.caching_checkbox.isChecked(),
                'searcher_enabled': self.searcher_checkbox.isChecked(),
            }
            initial_state = {
                'caching_enabled': settings["metadatastore"]["caching_enabled"],
                'searcher_enabled': settings["metadatastore"]["searcher_enabled"],
            }
            return current_state != initial_state
        except (KeyError, TypeError):
            # If settings don't exist, check against defaults
            return (self.caching_checkbox.isChecked() != True or
                    self.searcher_checkbox.isChecked() != True)

    def set_settings(self):
        """Save metadatastore settings."""
        if "metadatastore" not in settings.global_config:
            settings.global_config["metadatastore"] = {}

        settings.global_config["metadatastore"]["caching_enabled"] = self.caching_checkbox.isChecked()
        settings.global_config["metadatastore"]["searcher_enabled"] = self.searcher_checkbox.isChecked()

        logger.info(
            f"Metadatastore settings saved: "
            f"caching={self.caching_checkbox.isChecked()}, "
            f"searcher={self.searcher_checkbox.isChecked()}"
        )




"""Tests for MDSSearcher cross-database functionality."""
import pytest
import pandas as pd
from activity_browser.bwutils.metadata.searcher import MDSSearcher
from activity_browser.bwutils.metadata.metadata import MetaDataStore


@pytest.fixture
def multi_db_mds():
    """Create a MetaDataStore with multiple databases."""
    test_data = pd.DataFrame([
        [1, "db1", "coal production", "coal", "process", "", "", "", ""],
        [2, "db1", "coal mining", "coal", "process", "", "", "", ""],
        [3, "db1", "steel production", "steel", "process", "", "", "", ""],
        [4, "db2", "coal transport", "transport", "process", "", "", "", ""],
        [5, "db2", "electricity from coal", "electricity", "process", "", "", "", ""],
        [6, "db3", "coal combustion", "heat", "process", "", "", "", ""],
        [7, "db3", "gas production", "gas", "process", "", "", "", ""],
    ], columns=["id", "database", "name", "reference product", "type", "location", "unit", "comment", "tags"])

    mds = MetaDataStore()
    mds.dataframe = test_data
    return mds


def test_search_single_database(multi_db_mds):
    """Test searching within a single database."""
    searcher = MDSSearcher(multi_db_mds)

    # Search for "coal" in db1
    results = searcher.search("coal", database="db1")
    assert len(results) == 2
    assert set(results) == {1, 2}

    # Search for "coal" in db2
    results = searcher.search("coal", database="db2")
    assert len(results) == 2
    assert set(results) == {4, 5}

    # Search for "coal" in db3
    results = searcher.search("coal", database="db3")
    assert len(results) == 1
    assert set(results) == {6}


def test_search_all_databases(multi_db_mds):
    """Test searching across all databases when database=None."""
    searcher = MDSSearcher(multi_db_mds)

    # Search for "coal" across all databases
    results = searcher.search("coal", database=None)
    assert len(results) == 5
    assert set(results) == {1, 2, 4, 5, 6}

    # Search for "production" across all databases
    results = searcher.search("production", database=None)
    assert len(results) == 3
    assert set(results) == {1, 3, 7}


def test_fuzzy_search_all_databases(multi_db_mds):
    """Test fuzzy search across all databases."""
    searcher = MDSSearcher(multi_db_mds)

    # Fuzzy search for "coal" across all databases
    results = searcher.fuzzy_search("coal", database=None)
    assert len(results) >= 5

    # Fuzzy search for "production" across all databases
    results = searcher.fuzzy_search("production", database=None)
    assert len(results) >= 3


def test_search_cache_separation(multi_db_mds):
    """Test that search cache properly separates single-db and all-db searches."""
    searcher = MDSSearcher(multi_db_mds)

    # Do searches to populate cache
    results_db1 = searcher.search("coal", database="db1")
    results_all = searcher.search("coal", database=None)

    # Verify results are different
    assert len(results_db1) == 2
    assert len(results_all) == 5
    assert set(results_db1).issubset(set(results_all))

    # Search again to use cached results
    results_3ached = searcher.search("coal", database="db1")
    results_all_cached = searcher.search("coal", database=None)

    # Verify cached results match original
    assert results_db1 == results_3ached
    assert results_all == results_all_cached


def test_auto_complete_all_databases(multi_db_mds):
    """Test autocomplete across all databases."""
    searcher = MDSSearcher(multi_db_mds)

    # Autocomplete for "coa" across all databases
    completions = searcher.auto_complete("coa", database=None)
    assert "coal" in completions

    # Autocomplete for "prod" in specific database
    completions_db1 = searcher.auto_complete("prod", database="db1")
    assert "production" in completions_db1

    # Autocomplete for "prod" across all databases
    completions_all = searcher.auto_complete("prod", database=None)
    assert "production" in completions_all


def test_empty_search_all_databases(multi_db_mds):
    """Test empty search returns all items when database=None."""
    searcher = MDSSearcher(multi_db_mds)

    results = searcher.search("", database=None)
    assert len(results) == 7  # All items in all databases




from PySide6.QtCore import QModelIndex
from loguru import logger
from typing import Literal

from qtpy import QtWidgets, QtGui, QtCore
from qtpy.QtCore import Qt

import pandas as pd
import bw2data as bd

import bw_functional as bf

from activity_browser import app
from activity_browser.bwutils.commontasks import (refresh_node, database_is_locked, database_is_legacy,
                                                  is_node_product_or_waste, is_node_biosphere, parameters_in_scope,
                                                  is_node_product, is_node_waste)
from activity_browser.ui import widgets, icons, delegates, core



EXCHANGE_MAP = {
    "natural resource": "biosphere", "emission": "biosphere", "inventory indicator": "biosphere",
    "economic": "biosphere", "social": "biosphere", "product": "technosphere",
    "processwithreferenceproduct": "technosphere", "waste": "technosphere",
}


class ExchangesTab(QtWidgets.QWidget):
    """
    A widget that displays exchanges related to a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display exchanges for.
        output_view (ExchangesView): The view displaying the output exchanges.
        output_model (ExchangesModel): The model containing the data for the output exchanges.
        input_view (ExchangesView): The view displaying the input exchanges.
        input_model (ExchangesModel): The model containing the data for the input exchanges.
    """
    def __init__(self, activity: tuple | int | bd.Node, parent=None):
        """
        Initializes the ExchangesTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display exchanges for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.setAcceptDrops(True)

        # Refresh the activity node
        self.activity = refresh_node(activity)

        # Output Table
        self.output_view = ExchangesView(self)
        self.output_model = ExchangesModel(tab=self)
        self.output_view.setModel(self.output_model)

        # Set indentation for output view
        self.output_view.setIndentation(0)

        # Input Table
        self.input_view = ExchangesView(self)
        self.input_model = ExchangesModel(tab=self)
        self.input_view.setModel(self.input_model)

        # Set indentation for input view
        self.input_view.setIndentation(0)

        # Overlay for drag and drop
        self.overlay = None

        # Build the layout of the widget
        self.build_layout()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        # Add output label and view to the layout
        output = QtWidgets.QWidget(self)
        output_layout = QtWidgets.QVBoxLayout(output)
        output_layout.addWidget(widgets.ABLabel.demiBold(" Output:", self))
        output_layout.addWidget(self.output_view)

        # Add input label and view to the layout
        input = QtWidgets.QWidget(self)
        input_layout = QtWidgets.QVBoxLayout(input)
        input_layout.addWidget(widgets.ABLabel.demiBold(" Input:", self))
        input_layout.addWidget(self.input_view)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 10, 0, 1)
        splitter = QtWidgets.QSplitter(Qt.Orientation.Vertical, self, childrenCollapsible=False)
        splitter.addWidget(output)
        splitter.addWidget(input)
        layout.addWidget(splitter)

    def sync(self) -> None:
        """
        Synchronizes the widget with the current state of the activity.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        # Refresh the activity node
        self.activity = refresh_node(self.activity)

        # Get the production, technosphere, and biosphere exchanges
        production = self.activity.production()
        technosphere = self.activity.technosphere()
        biosphere = self.activity.biosphere()
        substitution = self.activity.substitution()

        # Filter inputs and outputs based on the amount and type
        inputs = ([x for x in production if x["amount"] < 0] +
                  [x for x in technosphere if x["amount"] >= 0] +
                  [x for x in biosphere if (x.input["type"] != "emission" and x["amount"] >= 0) or (x.input["type"] == "emission" and x["amount"] < 0)] +
                  [x for x in substitution if x["amount"] < 0]
                  )

        outputs = ([x for x in production if x["amount"] >= 0] +
                   [x for x in technosphere if x["amount"] < 0] +
                   [x for x in biosphere if (x.input["type"] == "emission" and x["amount"] >= 0) or (x.input["type"] != "emission" and x["amount"] < 0)] +
                   [x for x in substitution if x["amount"] >= 0]
                   )

        # Update the models with the new data
        output_df = self.build_df(outputs)
        output_df.reset_index(drop=True, inplace=True)
        self.output_model.set_dataframe(output_df)
        self.output_view.drag_drop_hint.setVisible(output_df.empty)

        input_df = self.build_df(inputs)
        input_df.reset_index(drop=True, inplace=True)
        self.input_model.set_dataframe(input_df)
        self.input_view.drag_drop_hint.setVisible(input_df.empty)

    def build_df(self, exchanges) -> pd.DataFrame:
        """
        Builds a DataFrame from the given exchanges.

        Args:
            exchanges (list): The list of exchanges to build the DataFrame from.

        Returns:
            pd.DataFrame: The DataFrame containing the exchanges data.
        """
        # Define the columns for the metadata
        cols = ["key", "unit", "name", "product", "location", "database", "allocation_factor",
                "properties", "processor", "categories", "type"]

        # Create a DataFrame from the exchanges
        exc_df = pd.DataFrame(exchanges, columns=["amount", "input", "formula", "comment", "type"])
        exc_df["uncertainty"] = [x.uncertainty for x in exchanges]
        act_df = app.metadata.get_metadata(exc_df["input"].unique(), cols).rename(columns={"type": "_producer_type"})

        # Merge the exchanges DataFrame with the metadata DataFrame
        df = exc_df.merge(
            act_df,
            left_on="input",
            right_on="key"
        ).drop(columns=["key"])

        # Set allocation_factor to NA for non-production exchanges
        df.loc[df["type"] != "production", "allocation_factor"] = pd.NA

        # Handle properties data if available
        if not act_df.properties.isna().all():
            props_df = act_df[act_df.properties.notna()]
            props_df = pd.DataFrame(list(props_df.get("properties")), index=props_df.key)
            props_df.rename(lambda col: f"property_{col}", axis="columns", inplace=True)

            df = df.merge(
                props_df,
                left_on="input",
                right_index=True,
                how="left",
            )

        # Add allocation and activity type information
        df["_allocate_by"] = self.activity.get("allocation")
        df["_activity_type"] = self.activity.get("type")
        df["_exchange"] = exchanges

        # Drop the properties column and rename some columns
        df.drop(columns=["properties"], inplace=True)
        df.rename({
            "input": "_input_key",
            "processor": "_processor_key",
            "type": "_exchange_type",
            "name": "producer",
        }, axis="columns", inplace=True)

        # Define the order of columns for the final DataFrame
        cols = ["amount", "unit", "product", "producer", "location", "categories", "database"]
        cols += ["allocation_factor"] if not database_is_legacy(self.activity.get("database")) else []
        cols += [col for col in df.columns if col.startswith("property")]
        cols += ["formula", "comment", "uncertainty"]
        cols += [col for col in df.columns if col.startswith("_")]

        return df[cols]

    def dragEnterEvent(self, event):
        """
        Handles the drag enter event.

        Args:
            event: The drag enter event.
        """
        if database_is_locked(self.activity["database"]):
            return

        has_nodes = event.mimeData().hasFormat("application/bw-nodekeylist")
        has_exchanges = event.mimeData().hasFormat("application/bw-exchangelist")

        if not has_nodes and not has_exchanges:
            return

        event.accept()
        action = self.action_from_mime(event.mimeData())

        self.input_view.overlay.show()
        self.output_view.overlay.show()

        if action == "product":
            self.output_view.overlay.setText("Drop to substitute production")
            self.input_view.overlay.setText("Drop to consume product")
            return

        if action == "waste":
            self.output_view.overlay.setText("Drop to produce waste")
            self.input_view.overlay.setText("Drop to substitute waste consumption")
            return

        if action == "resource":
            self.output_view.overlay.hide()
            self.input_view.overlay.setText("Drop to consume natural resource")
            return

        if action == "emission":
            self.input_view.overlay.hide()
            self.output_view.overlay.setText("Drop to emit to environment")
            return


    def dragMoveEvent(self, event):
        """
        Handles the drag move event to adjust overlay opacity based on hover position.

        Args:
            event: The drag move event.
        """
        has_nodes = event.mimeData().hasFormat("application/bw-nodekeylist")
        has_exchanges = event.mimeData().hasFormat("application/bw-exchangelist")

        if not has_nodes and not has_exchanges:
            return

        if self.input_view.overlay.hovering():
            self.input_view.overlay.setOpacity("high")
            self.output_view.overlay.setOpacity("medium")
        elif self.output_view.overlay.hovering():
            self.output_view.overlay.setOpacity("high")
            self.input_view.overlay.setOpacity("medium")
        else:
            self.input_view.overlay.setOpacity("medium")
            self.output_view.overlay.setOpacity("medium")
            event.ignore()
            return

        event.accept()

    def dragLeaveEvent(self, event):
        """
        Handles the drag leave event.

        Args:
            event: The drag leave event.
        """
        # Reset the palette on drag leave
        self.input_view.overlay.hide()
        self.output_view.overlay.hide()

    def dropEvent(self, event):
        """
        Handles the drop event.

        Args:
            event: The drop event.
        """
        logger.debug(f"Dropevent from: {type(event.source()).__name__} to: {self.__class__.__name__}")

        self.input_view.overlay.hide()
        self.output_view.overlay.hide()

        output = self.output_view.overlay.hovering()
        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")

        positive_exchanges = {"technosphere": set(), "biosphere": set(), "substitution": set()}
        negative_exchanges = {"technosphere": set(), "substitution": set()}

        for key in keys:
            exc_type = get_exchange_type(key, output=output)
            if exc_type is None:
                continue
            if exc_type.startswith("-"):
                negative_exchanges[exc_type[1:]].add(key)
            else:
                positive_exchanges[exc_type].add(key)

        # Run the action for new exchanges
        for exc_type, keys in positive_exchanges.items():
            app.actions.ExchangeNew.run(keys, self.activity.key, exc_type)
        for exc_type, keys in negative_exchanges.items():
            app.actions.ExchangeNew.run(keys, self.activity.key, exc_type, amount=-1)

    def action_from_mime(self, mime: core.ABMimeData) -> Literal["product", "waste", "resource", "emission", "generic"]:
        """
        Determines the appropriate action based on the mime data.

        Args:
            mime (core.ABMimeData): The mime data.

        """
        keys = mime.retrievePickleData("application/bw-nodekeylist")
        data = app.metadata.get_metadata(keys, ["type"])
        data = set(data["type"].unique())
        data.discard("process")
        data.discard("multifunctional")
        data.discard("nonfunctional")

        if len(data) != 1:
            return "generic"

        node_type = data.pop()
        if node_type in ["product", "processwithreferenceproduct"]:
            return "product"
        if node_type == "waste":
            return "waste"
        if node_type == "natural resource":
            return "resource"
        if node_type == "emission":
            return "emission"
        else:
            return "generic"

def get_exchange_type(activity_key: tuple, output=False) -> str | None:
    if is_node_product(activity_key):
        return "substitution" if output else "technosphere"
    if is_node_waste(activity_key):
        return "-technosphere" if output else "-substitution"
    elif is_node_biosphere(activity_key):
        return "biosphere"
    return None


class RelinkDelegate(delegates.StringDelegate):
    matched: pd.DataFrame

    def createEditor(self, parent, option, index):
        model: ExchangesModel = index.model()
        
        column = model.column_name(index)
        column = "name" if column == "producer" else column

        if column == "product" and model.functional(index):
            return super().createEditor(parent, option, index)
        
        row = model.row(index)

        setup = {
            "database": row["database"],
            "name": row["producer"],
            "product": row["product"],
            "categories": row["categories"],
            "location": row["location"],
            "type": row["_producer_type"],
        }

        del setup[column]  # remove the column being edited because we are looking for alternatives

        self.matched = app.metadata.match(**setup)

        combo = QtWidgets.QComboBox(parent)
        combo.addItems(list(self.matched.get(column, []).astype(str)))
        return combo

    def setEditorData(self, editor: QtWidgets.QComboBox, index):
        model: ExchangesModel = index.model()
        column = model.column_name(index)
        column = "name" if column == "producer" else column

        if column == "product" and model.functional(index):
            return super().setEditorData(editor, index)

        value = index.data()
        if value:
            i = editor.findText(str(value))
            if i >= 0:
                editor.setCurrentIndex(i)

    def setModelData(self, editor: QtWidgets.QComboBox, model, index):
        model: ExchangesModel = index.model()
        column = model.column_name(index)
        column = "name" if column == "producer" else column

        if column == "product" and model.functional(index):
            return super().setModelData(editor, model, index)

        choice = editor.currentIndex()
        key = self.matched.iloc[choice].key
        row = model.row(index)

        app.actions.ExchangeModify.run(
            row.get("_exchange"),
            {"input": key}
        )


class ExchangesView(widgets.ABTreeView):
    """
    A view that displays the exchanges in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
        hovered_item (ExchangesItem): The item currently being hovered over.
    """
    defaultColumnDelegates = {
        "amount": delegates.AbsoluteAmountDelegate,
        "allocation_factor": delegates.FloatDelegate,
        "substitution_factor": delegates.FloatDelegate,
        "unit": delegates.StringDelegate,
        "producer": RelinkDelegate,
        "location": RelinkDelegate,
        "product": RelinkDelegate,
        "database": RelinkDelegate,
        "categories": RelinkDelegate,
        "formula": delegates.NewFormulaDelegate,
        "comment": delegates.StringDelegate,
        "uncertainty": delegates.UncertaintyDelegate,
    }

    class HeaderMenu(widgets.ABMenu):
        menuSetup = [
            lambda m: m.setup_view_menu(),
            lambda m: m.setup_allocation(),
        ]

        def setup_view_menu(self):
            table_view: ExchangesView = self.parent()
            table_model: ExchangesModel = table_view.model()

            def toggle_slot(action: QtWidgets.QAction):
                """
                Toggles the visibility of columns based on the action triggered.

                Args:
                    action (QtWidgets.QAction): The action triggered.
                """
                indices = action.data()
                for index in indices:
                    hidden = table_view.isColumnHidden(index)
                    table_view.setColumnHidden(index, not hidden)

            # Create the view menu
            view_menu = QtWidgets.QMenu(table_view)
            view_menu.setTitle("View")

            props_indices = []

            # Add actions for each column to the view menu
            for i, col in enumerate(table_model.columns()):
                if col.startswith("property"):
                    props_indices.append(i)
                    continue

                action = QtWidgets.QAction(table_model.columns()[i], self)
                action.setCheckable(True)
                action.setChecked(not table_view.isColumnHidden(i))
                action.setData([i])

                view_menu.addAction(action)

            # Add a combined action for property columns
            if props_indices:
                action = QtWidgets.QAction("properties", self)
                action.setCheckable(True)
                action.setChecked(not table_view.isColumnHidden(props_indices[0]))
                action.setData(props_indices)
                view_menu.addAction(action)

            # Connect the view menu actions to the toggle slot
            view_menu.triggered.connect(toggle_slot)

            # Add the view menu to the context menu
            self.addMenu(view_menu)

        def setup_allocation(self):
            table_view: ExchangesView = self.parent()

            if database_is_locked(table_view.activity["database"]) or not self.column.startswith("property"):
                return

            action = app.actions.ActivityModify.get_QAction(table_view.activity.key,
                                                        "allocation",
                                                        self.column[9:],
                                                        parent=self)
            action.setText(f"Allocate by {self.column[9:]}")
            self.addAction(action)

        @property
        def column(self):
            view, model, pos = self.parent(), self.parent().model(), QtGui.QCursor.pos()
            col_index = view.columnAt(view.mapFromGlobal(pos).x())
            return model.columns()[col_index]

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m: m.add(app.actions.ActivityNewProduct, [m.activity.key],
                            enable=not m.locked and not database_is_legacy(m.activity["database"])
                            ),
            lambda m: m.add(app.actions.ActivityNewProduct, [m.activity.key], "waste",
                            enable=not m.locked and not database_is_legacy(m.activity["database"]),
                            text="Create waste"
                            ),
            lambda m: m.addSeparator(),
            lambda m: m.add(app.actions.ExchangeDelete, m.exchanges, enable=bool(m.exchanges) and not m.locked),
            lambda m: m.add(app.actions.ExchangeSDFToClipboard, m.exchanges, enable=bool(m.exchanges)),
            lambda m: m.add(app.actions.ActivityOpen, [x.input for x in m.exchanges],
                            enable=bool(m.exchanges),
                            text="Open processs" if len(m.exchanges) == 1 else "Open processes",
                            ),
        ]

        @property
        def locked(self):
            return database_is_locked(self.activity["database"])

        @property
        def activity(self):
            return self.parent().activity

        @property
        def exchanges(self):
            indexes = self.parent().selectedIndexes()
            exchanges = [i.model().get(i, "_exchange") for i in indexes]
            return list(set(exchanges))

    def __init__(self, parent):
        """
        Initializes the ExchangesView.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
        """
        super().__init__(parent)
        self.setSortingEnabled(True)

        # Enable drag and drop
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)

        self.drag_drop_hint = QtWidgets.QLabel("Drag products here to create new exchanges.", self)
        fnt = self.drag_drop_hint.font()
        fnt.setPointSize(fnt.pointSize() + 2)
        fnt.setWeight(QtGui.QFont.Weight.ExtraLight)
        self.drag_drop_hint.setFont(fnt)

        # Set up the layout
        layout = QtWidgets.QVBoxLayout(self)
        layout.addStretch()
        layout.addWidget(self.drag_drop_hint, alignment=Qt.AlignmentFlag.AlignCenter)  # Center horizontally
        layout.addStretch()

        # Set the property delegate
        self.propertyDelegate = delegates.PropertyDelegate(self)
        self.overlay = widgets.ABDropOverlay(self)
        self.overlay.hide()

    @property
    def activity(self):
        """
        Returns the activity associated with the view.

        Returns:
            The activity associated with the view.
        """
        return self.parent().parent().parent().activity

    def setDefaultColumnDelegates(self):
        """
        Sets the default column delegates for the view.
        """
        super().setDefaultColumnDelegates()

        columns = self.model().columns()
        for i, col_name in enumerate(columns):
            if not col_name.startswith("property_"):
                continue
            # Set the delegate for property columns
            self.setItemDelegateForColumn(i, self.propertyDelegate)

    def startDrag(self, supportedActions: Qt.DropAction) -> None:
        """
        Initiates a drag operation with the selected exchanges.

        Args:
            supportedActions: The supported drop actions.
        """
        if database_is_locked(self.activity["database"]):
            return

        super().startDrag(supportedActions)


class ExchangesModel(core.ABTreeModel):
    """
    A model representing the data for the exchanges.
    """
    def __init__(self, tab: ExchangesTab):
        super().__init__(parent=tab, enable_sorting=True)
        self.tab = tab

    def mimeTypes(self) -> list[str]:
        """
        Returns the list of MIME types that this model supports.

        Returns:
            list[str]: List of supported MIME types.
        """
        return ["application/bw-exchangelist"]

    def mimeData(self, indices: list[QtCore.QModelIndex]) -> core.ABMimeData:
        """
        Returns the MIME data for the given indices.

        Args:
            indices (list[QtCore.QModelIndex]): The indices to get the MIME data for.

        Returns:
            core.ABMimeData: The MIME data containing the exchanges.
        """
        data = core.ABMimeData()
        exchanges = [self.get(index, "_exchange") for index in indices if index.isValid() and index.column() == 0]
        exchanges = [exc for exc in exchanges if exc is not None]
        data.setPickleData("application/bw-exchangelist", exchanges)
        return data

    def setData(self, index: QtCore.QModelIndex, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        """
        Sets the data for the given index.

        Args:
            index (QtCore.QModelIndex): The index to set data for.
            value: The value to set.
            role (int): The role for which to set the data.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if role != Qt.ItemDataRole.EditRole:
            return False

        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return False

        exchange = row.get("_exchange")
        if exchange is None:
            return False

        if column_name in ["amount", "formula", "comment"]:
            if column_name == "formula" and not str(value).strip():
                app.actions.ExchangeFormulaRemove.run([exchange])
                return True

            app.actions.ExchangeModify.run(exchange, {column_name.lower(): value})
            return True

        if column_name in ["unit", "product", "location", "substitution_factor", "allocation_factor"]:
            act = exchange.input
            app.actions.ActivityModify.run(act.key, column_name.lower(), value)
            return True

        if column_name.startswith("property_"):
            # should move this process to a separate action
            process = exchange.output
            product = exchange.input

            if not isinstance(process, bf.Process) or not isinstance(product, bf.Product):
                logger.warning(f"Expected a Process and Product, got {type(process)} and {type(product)} instead.")
                return False

            prop_key = column_name[9:]

            prop = process.property_template(prop_key, value)

            props = product.get("properties", {})
            props[prop_key] = prop

            app.actions.ActivityModify.run(product, "properties", props)
            return True

        return False
    
    def decorationData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides decoration data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide decoration data.

        Returns:
            The decoration data for the index.
        """
        column_name = self.column_name(index)

        if column_name in ["product", "producer"]:
            activity_type = self.get(index, "_producer_type")
            if activity_type in ["natural resource", "emission", "inventory indicator", "economic", "social"]:
                return icons.qicons.biosphere if column_name == "producer" else None
            if activity_type == "processwithreferenceproduct":
                return icons.qicons.processproduct if column_name == "producer" else icons.qicons.product
            if activity_type in ["product", "process", "multifunctional", "nonfunctional"]:
                return icons.qicons.process if column_name == "producer" else icons.qicons.product
            if activity_type == "waste":
                return icons.qicons.process if column_name == "producer" else icons.qicons.waste

        if column_name == "amount":
            formula = self.get(index, "formula")
            if pd.isna(formula) or formula is None or formula == "":
                return None
            return icons.qicons.parameterized

        return None
    
    def fontData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides font data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide font data.

        Returns:
            QtGui.QFont: The font data for the index.
        """
        if self.substituted(index):
            font = QtGui.QFont()
            font.setItalic(True)
            font.setWeight(QtGui.QFont.Weight.DemiBold)
            return font

        if self.functional(index):
            font = QtGui.QFont()
            font.setWeight(QtGui.QFont.Weight.DemiBold)
            return font

        return None

    def indexEditable(self, index):
        column_name = self.column_name(index)
        database = self.get(index, "_exchange")["output"][0]

        # Prevent editing if the database is locked
        if database_is_locked(database):
            return False
        
        functional = self.functional(index)

        # Allow editing for specific keys: "amount", "formula", and "uncertainty".
        if column_name in ["amount", "formula", "uncertainty", "comment"]:
            return True

        # Allow editing for "unit", "name", and "substitution_factor" if the exchange is functional.
        if column_name in ["unit", "product"] and functional:
            return True

        # Allow editing for "producer", "location", "categories", and "database" if the exchange is not functional.
        if column_name in ["producer", "product", "location", "categories", "database"] and not functional:
            return True

        # Allow editing for properties (keys starting with "property_") if the exchange is functional.
        if column_name.startswith("property_") and functional:
            return True
        
        # Allow editing for allocation_factor if functional and allocation is manual
        if column_name == "allocation_factor" and functional and self.tab.activity.get("allocation") == "manual":
            return True

        return False

    def indexDragEnabled(self, index: QModelIndex) -> bool:
        return True
    
    def functional(self, index):
        """
        Returns whether the index is functional.

        Args:
            index (QtCore.QModelIndex): The index to check.

        Returns:
            bool: True if the index is functional, False otherwise.
        """
        return self.get(index, "_exchange_type") == "production"

    def substituted(self, index):
        """
        Returns whether the index is functional.

        Args:
            index (QtCore.QModelIndex): The index to check.

        Returns:
            bool: True if the index is functional, False otherwise.
        """
        return self.get(index, "_exchange_type") == "substitution"
    
    def scoped_parameters(self, index):
        """
        Returns the scoped parameters for the index.

        Args:
            index (QtCore.QModelIndex): The index to get scoped parameters for.

        Returns:
            list: A list of scoped parameters for the index.
        """
        exchange = self.get(index, "_exchange")
        return parameters_in_scope(exchange.output)
    


from .plot import ABPlot
from .abstract_pane import ABAbstractPane
from .comparison_switch import SwitchComboBox
from .cutoff_menu import CutoffMenu
from .line_edit import (ABLineEdit, SignalledComboEdit, SignalledLineEdit,
                        SignalledPlainTextEdit)
from .text_edit import (ABAutoCompleTextEdit, ABTextEdit, MetaDataAutoCompleteTextEdit)
from .line import ABHLine, ABVLine
from .formula_edit import ABFormulaEdit

from .combobox import ABComboBox, CheckableComboBox
from .button_collapser import ABRadioButtonCollapser
from .wizard import ABWizard
from .wizard_page import ABWizardPage, ABThreadedWizardPage
from .file_selector import ABFileSelector
from .database_name_edit import DatabaseNameEdit
from .dock_widget import ABDockWidget
from .label import ABLabel
from .central import CentralTabWidget
from .menu import ABMenu
from .drop_overlay import ABDropOverlay
from .tree_view import ABTreeView
from .buttons import ABCloseButton, ABMinimizeButton
from .tab_widget import ABTabWidget
from .web_engine_page import ABWebEnginePage
from .abstract_navigator import ABAbstractNavigator, ABAbstractGraph



import os

import pandas as pd

from qtpy import QtWidgets

from activity_browser.app import application
from activity_browser.app.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils.utils import Parameters


class SaveParametersToExcel(ABAction):
    """
    ABAction to export database(s) to Excel format (.xlsx).
    """
    text = "Save parameters to Excel (.xlsx)"
    tool_tip = "Save parameters to Excel format"

    @classmethod
    @exception_dialogs
    def run(cls, file_path: str = None):
        if file_path is None:
            suggestion = os.path.expanduser("~/parameters.xlsx")

            file_path, _ = QtWidgets.QFileDialog.getSaveFileName(
                parent=application.main_window,
                caption=f'Export parameters to Excel',
                dir=suggestion,
                filter='Excel spreadsheet (*.xlsx);; All files (*.*)'
            )

        if not file_path:
            return

        data = [p[:3] for p in Parameters.from_bw_parameters()]
        df = pd.DataFrame(data, columns=["Name", "Group", "default"]).set_index("Name")
        df.to_excel(file_path)

        os.startfile(file_path)



from typing import Literal, Optional
from loguru import logger

from qtpy.QtCore import QObject

import pandas as pd

from activity_browser.bwutils.settings import Settings
from .fields import all_fields, all_types


class MetaDataStore(QObject):
    """Singleton class to manage metadata storage, loading, updating, and searching."""
    _instance = None
    
    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance._initialized = False
        return cls._instance
    
    def __init__(self, parent=None):
        from .loader import MDSLoader
        from .updater import MDSUpdater
        from .searcher import MDSSearcher

        if self._initialized:
            return
        self._initialized = True
        super().__init__(parent=parent)

        self._dataframe = pd.DataFrame()

        self._added: set[tuple[str, str]] = set()
        self._updated: set[tuple[str, str]] = set()
        self._deleted: set[tuple[str, str]] = set()

        self.loader = MDSLoader(self)
        self.updater = MDSUpdater(self)
        self.searcher: MDSSearcher | None = None  # initialized by the loader

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._dataframe

    @dataframe.setter
    def dataframe(self, df: pd.DataFrame) -> None:
        # Ensure all expected columns are present, in the correct order
        df = df.reindex(columns=all_fields)[all_fields]

        # Apply types carefully - avoid in-place modifications
        for col, col_type in all_types.items():
            if col in df.columns:
                df[col] = df[col].astype(col_type)

        # No NaN values in object columns, use None instead
        for col, col_type in all_types.items():
            if col_type != object or col not in df.columns:
                continue
            df[col] = df[col].where(df[col].notnull(), None)

        self._dataframe = df

    @property
    def databases(self):
        return set(self._dataframe.index.get_level_values(0).unique().tolist())

    @property
    def keys(self):
        return set(self._dataframe.index.tolist())

    def register_mutation(self, key: tuple[str, str], action: Literal["add", "update", "delete"]):
        if action == "add":
            self._added.add(key)
            self._updated.discard(key)
            self._deleted.discard(key)

        elif action == "update":
            if key not in self._added:
                self._updated.add(key)

        elif action == "delete":
            if key in self._added:
                self._added.discard(key)
            else:
                self._deleted.add(key)
            self._updated.discard(key)
        else:
            raise ValueError(f"Unknown action: {action}")

    def flush_mutations(self) -> tuple[set[tuple[str, str]], set[tuple[str, str]], set[tuple[str, str]]]:
        from activity_browser.bwutils import filesystem

        if not (self._added or self._updated or self._deleted):
            return set(), set(), set()

        added = self._added.copy()
        updated = self._updated.copy()
        deleted = self._deleted.copy()

        self._added.clear()
        self._updated.clear()
        self._deleted.clear()

        if Settings()["metadatastore"]["caching_enabled"]:
            cache_path = filesystem.get_project_ab_path() / "metadatastore_cache.pkl"
            self._dataframe.to_pickle(cache_path)

        return added, updated, deleted

    def match(self, **kwargs: dict[str, str]) -> pd.DataFrame:
        """Return a slice of the dataframe matching the criteria.
        """
        df = self._dataframe.query(
            " and ".join(
                [
                    f"`{key}`.astype('str') == {str(value)!r}" if not pd.isna(value) else f"`{key}`.isnull()"
                    for key, value in kwargs.items()
                ])
        )

        return df

    def get_metadata(self, keys: list = None, columns: list = None) -> pd.DataFrame:
        """Return a slice of the dataframe matching row and column identifiers.

        NOTE: https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#deprecate-loc-reindex-listlike
        From pandas version 1.0 and onwards, attempting to select a column
        with all NaN values will fail with a KeyError.
        """
        keys = keys if keys is not None else self._dataframe.index.tolist()
        columns = columns if columns is not None else all_fields

        df = self._dataframe.loc[pd.IndexSlice[keys], :]
        return df.reindex(columns, axis="columns")

    def get_database_metadata(self, db_name: str, columns: list = None) -> pd.DataFrame:
        columns = columns if columns is not None else all_fields

        if db_name not in self.databases:
            return pd.DataFrame(columns=columns or all_fields)

        df = self._dataframe.loc[[db_name], columns]
        return df.reindex(columns, axis="columns")

    def _pandas_search(self, query: str, database: str = None, columns: list = None) -> pd.DataFrame:
        """Fallback pandas-based search when searcher is not initialized.

        Args:
            query: Search query string, may contain key:value parameters
            database: Optional database name to restrict search
            columns: Optional list of columns to return

        Returns:
            DataFrame with matching results
        """
        params, clean_query = get_query_parameters(query)
        columns = columns if columns is not None else all_fields

        # Start with the full dataframe or database subset
        if database and database in self.databases:
            df = self._dataframe.loc[[database]]
        else:
            df = self._dataframe

        if not clean_query.strip():
            # If no search query, just filter by parameters
            if params:
                extra_query = " & ".join(
                    [
                        f"`{key}`.astype('str').str.contains('{value}', case=False)"
                        for key, value in params.items()
                        if key in df.columns
                    ]
                )
                if extra_query:
                    df = df.query(extra_query)
            return df[columns]

        # Search across text fields: name, product, synonyms, categories, unit, location
        search_fields = ['name', 'product', 'synonyms', 'categories', 'unit', 'location', 'CAS number']
        mask = pd.Series([False] * len(df), index=df.index)

        for field in search_fields:
            if field in df.columns:
                # Case-insensitive search
                mask |= df[field].astype(str).str.contains(clean_query, case=False, na=False)

        df = df[mask]

        # Apply additional parameter filters if any
        if params:
            extra_query = " & ".join(
                [
                    f"`{key}`.astype('str').str.contains('{value}', case=False)"
                    for key, value in params.items()
                    if key in df.columns
                ]
            )
            if extra_query:
                df = df.query(extra_query)

        return df[columns] if columns else df

    def search(self, query: str, columns: list = None) -> pd.DataFrame:
        if self.searcher:
            # Advanced searcher is initialized, so use that
            params, query = get_query_parameters(query)
            result = self.searcher.search(query)
            return self._meta_from_result(params, result, columns)

        # Fallback to simple pandas search
        logger.debug("Using simple pandas search as searcher is not initialized.")
        return self._pandas_search(query, columns=columns)

    def search_database(self, query: str, database: str, columns: list = None) -> pd.DataFrame:
        if self.searcher:
            params, query = get_query_parameters(query)
            result = self.searcher.fuzzy_search(query, database=database)
            return self._meta_from_result(params, result, columns)

        # Fallback to simple pandas search
        logger.debug(f"Using simple pandas search for database '{database}' as searcher is not initialized.")
        return self._pandas_search(query, database=database, columns=columns)

    def _meta_from_result(self, params: dict, result: list[int], columns: list = None) -> pd.DataFrame:
        df = self._dataframe.loc[self.dataframe["id"].isin(result), columns or all_fields]
        df.sort_values(by="id", inplace=True, key=lambda x: x.map({id_: i for i, id_ in enumerate(result)}))

        extra_query = " & ".join(
            [
                f"`{key}`.astype('str').str.contains('{value}', False)"
                for key, value in params.items()
                if key in df.columns
            ]
        )
        if extra_query:
            df = df.query(extra_query)

        return df

    def auto_complete(self, word: str, context: Optional[set] = None, database: Optional[str] = None):
        if not self.searcher:
            logger.warning(f"Attempted to search metadata before searcher was initialized.")
            return []

        word = self.searcher.clean_text(word)
        completions = self.searcher.auto_complete(word, context=context, database=database)
        return completions

    def clear_cache(self):
        from activity_browser.bwutils import filesystem

        cache_path = filesystem.get_project_ab_path() / "metadatastore_cache.pkl"
        if cache_path.exists():
            cache_path.unlink()
            logger.info("Metadata store cache cleared.")
        else:
            logger.info("No metadata store cache found to clear.")


def get_query_parameters(query: str) -> tuple[dict[str, str], str]:
    """Extract key-value pairs from a query string of the form 'key1:value1 key2:value2'."""
    params = {}
    tokens = query.split()
    clean_query = []
    for token in tokens:
        if ':' in token:
            key, value = token.split(':', 1)
            params[key] = value
        else:
            clean_query.append(token)
    return params, ' '.join(clean_query)



from qtpy import QtWidgets, QtCore

from activity_browser.ui import widgets

from .parameters_section import ParametersSection
from .parameterized_exchanges_section import ParameterizedExchangesSection


class ParametersPage(QtWidgets.QWidget):
    """
    A widget that displays all parameters and parameterized exchanges in the current project.

    This page shows:
    - Parameters section: A tree view of parameters organized by scope
    - Parameterized exchanges section: A table of exchanges with formulas
    """

    def __init__(self, parent=None):
        """
        Initializes the ParametersPage widget.

        Args:
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)

        self.parameters_section = ParametersSection(self)
        self.parameterized_exchanges_section = ParameterizedExchangesSection(self)

        self.build_layout()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 3, 0, 0)

        # Add both sections in a splitter
        splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical, self)

        # Parameters section
        params_widget = QtWidgets.QWidget()
        params_layout = QtWidgets.QVBoxLayout(params_widget)
        params_layout.setContentsMargins(0, 0, 0, 0)
        params_label = widgets.ABLabel.demiBold("    Parameters")
        params_layout.addWidget(params_label)
        params_layout.addWidget(widgets.ABHLine(self))
        params_layout.addWidget(self.parameters_section)
        splitter.addWidget(params_widget)

        # Parameterized exchanges section
        exchanges_widget = QtWidgets.QWidget()
        exchanges_layout = QtWidgets.QVBoxLayout(exchanges_widget)
        exchanges_layout.setContentsMargins(0, 0, 0, 0)
        exchanges_label = widgets.ABLabel.demiBold("    Parameterized Exchanges")
        exchanges_layout.addWidget(exchanges_label)
        exchanges_layout.addWidget(widgets.ABHLine(self))
        exchanges_layout.addWidget(self.parameterized_exchanges_section)
        splitter.addWidget(exchanges_widget)

        layout.addWidget(splitter)
        self.setLayout(layout)





from __future__ import annotations

from loguru import logger
from typing import Optional, Tuple

import numpy as np
import seaborn as sns

from qtpy import QtCore, QtGui, QtWidgets
import stats_arrays as sa

from activity_browser.ui.widgets import ABPlot




EMPTY_UNCERTAINTY = {
    "uncertainty type": sa.UndefinedUncertainty.id,
    "loc": np.NaN,
    "scale": np.NaN,
    "shape": np.NaN,
    "minimum": np.NaN,
    "maximum": np.NaN,
    "negative": False,
}


class UncertaintyDialog(QtWidgets.QDialog):
	"""Single-step dialog for defining a stats_arrays uncertainty.

	Mirrors the behavior of the UncertaintyWizard type page but returns a
	stats_arrays structured array on accept.

	Usage:
		ok, array = UncertaintyDialog.get_uncertainty(parent, initial=dict(...))
		if ok:
			# array is a numpy structured array compatible with stats_arrays
	"""

	def __init__(self, parent=None, initial: Optional[dict] = None):
		super().__init__(parent)
		self.setWindowTitle("Set Uncertainty")
		self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

		# State
		self.dist = None
		self.result_array = None  # Filled on accept
		self.result_dict = None  # Filled on accept
		self.previous_dist_id: Optional[int] = None
		self.mean_is_calculated = {
			sa.TriangularUncertainty.id,
			sa.UniformUncertainty.id,
			sa.DiscreteUniform.id,
			sa.BetaUncertainty.id,
		}

		# Top: distribution selection
		box1 = QtWidgets.QGroupBox("Select the uncertainty distribution")
		self.distribution = QtWidgets.QComboBox(box1)
		self.distribution.addItems([ud.description for ud in sa.uncertainty_choices])
		self.distribution.currentIndexChanged.connect(self._on_distribution_changed)

		header_layout = QtWidgets.QGridLayout()
		header_layout.addWidget(QtWidgets.QLabel("Distribution:"), 0, 0)
		header_layout.addWidget(self.distribution, 0, 1)
		box1.setLayout(header_layout)

		# Middle: parameters
		self.fields_box = QtWidgets.QGroupBox("Fill out required parameters")
		self.locale = QtCore.QLocale(
			QtCore.QLocale.English, QtCore.QLocale.UnitedStates
		)
		self.locale.setNumberOptions(QtCore.QLocale.RejectGroupSeparator)
		self.validator = QtGui.QDoubleValidator()
		self.validator.setLocale(self.locale)

		# loc/mean
		self.loc_label = QtWidgets.QLabel("Loc:")
		self.loc = QtWidgets.QLineEdit()
		self.loc.setValidator(self.validator)
		self.loc.textEdited.connect(self._sync_mean_from_loc)
		self.loc.textEdited.connect(self._check_negative)
		self.loc.textEdited.connect(self._generate_plot)

		self.mean_label = QtWidgets.QLabel("Mean:")
		self.mean = QtWidgets.QLineEdit()
		self.mean.setValidator(self.validator)
		self.mean.textEdited.connect(self._sync_loc_from_mean)
		self.mean.textEdited.connect(self._check_negative)
		self.mean.textEdited.connect(self._generate_plot)

		# Calculated mean (read-only) for some dists
		self.calc_mean_label = QtWidgets.QLabel("Mean:")
		self.calc_mean = QtWidgets.QLineEdit("nan")
		self.calc_mean.setDisabled(True)

		# Other parameters
		self.scale_label = QtWidgets.QLabel("Sigma/scale:")
		self.scale = QtWidgets.QLineEdit()
		self.scale.setValidator(self.validator)
		self.scale.textEdited.connect(self._generate_plot)

		self.shape_label = QtWidgets.QLabel("Shape:")
		self.shape = QtWidgets.QLineEdit()
		self.shape.setValidator(self.validator)
		self.shape.textEdited.connect(self._generate_plot)

		self.min_label = QtWidgets.QLabel("Minimum:")
		self.minimum = QtWidgets.QLineEdit()
		self.minimum.setValidator(self.validator)
		self.minimum.textEdited.connect(self._generate_plot)

		self.max_label = QtWidgets.QLabel("Maximum:")
		self.maximum = QtWidgets.QLineEdit()
		self.maximum.setValidator(self.validator)
		self.maximum.textEdited.connect(self._generate_plot)

		# Hidden flag for negative mean on lognormal
		self.negative = QtWidgets.QRadioButton(self)
		self.negative.setChecked(False)
		self.negative.setHidden(True)

		params_layout = QtWidgets.QGridLayout()
		# row 0: read-only calculated mean (will be hidden for most dists)
		params_layout.addWidget(self.calc_mean_label, 0, 0)
		params_layout.addWidget(self.calc_mean, 0, 1)
		# row 1: loc/mean pair
		params_layout.addWidget(self.loc_label, 1, 0)
		params_layout.addWidget(self.loc, 1, 1)
		params_layout.addWidget(self.mean_label, 1, 3)
		params_layout.addWidget(self.mean, 1, 4)
		# row 2+: other params
		params_layout.addWidget(self.scale_label, 2, 0)
		params_layout.addWidget(self.scale, 2, 1)
		params_layout.addWidget(self.shape_label, 3, 0)
		params_layout.addWidget(self.shape, 3, 1)
		params_layout.addWidget(self.min_label, 4, 0)
		params_layout.addWidget(self.minimum, 4, 1)
		params_layout.addWidget(self.max_label, 5, 0)
		params_layout.addWidget(self.maximum, 5, 1)
		self.fields_box.setLayout(params_layout)

		# Bottom: plot
		self.plot = SimpleDistributionPlot(self)

		# Buttons
		self.buttons = QtWidgets.QDialogButtonBox(
			QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
		)
		self.buttons.accepted.connect(self._on_accept)
		self.buttons.rejected.connect(self.reject)

		# Layout
		layout = QtWidgets.QVBoxLayout()
		layout.addWidget(box1)
		layout.addWidget(self.fields_box)
		layout.addWidget(self.plot)
		layout.addWidget(self.buttons)
		self.setLayout(layout)

		# Initialize values (defaults or provided initial)
		self._apply_initial(initial or {})
		self._on_distribution_changed(self.distribution.currentIndex())
		self._sync_mean_from_loc()
		self._generate_plot()

	# ---------- Public API ----------
	@staticmethod
	def get_uncertainty_array(
		parent=None, initial: Optional[dict] = None
	) -> Tuple[bool, Optional[np.ndarray]]:
		dlg = UncertaintyDialog(parent, initial=initial)
		ok = dlg.exec_() == QtWidgets.QDialog.Accepted
		return ok, dlg.result_array if ok else None
	
	@staticmethod
	def get_uncertainty_dict(
		parent=None, initial: Optional[dict] = None
	) -> Tuple[bool, Optional[dict]]:
		dlg = UncertaintyDialog(parent, initial=initial)
		ok = dlg.exec_() == QtWidgets.QDialog.Accepted
		return ok, dlg.result_dict if ok else None

	# ---------- Internal helpers ----------
	def _apply_initial(self, initial: dict) -> None:
		# Use EMPTY_UNCERTAINTY defaults, overridden by initial
		data = {k: v for k, v in EMPTY_UNCERTAINTY.items()}
		data.update(initial or {})
		# Distribution
		try:
			uc_type = int(data.get("uncertainty type", 0))
		except Exception:
			uc_type = 0
		self.distribution.setCurrentIndex(uc_type)
		# Fields (string form for QLineEdit)
		def to_str(val):
			return "nan" if val is None or (isinstance(val, float) and np.isnan(val)) else str(val)

		self.loc.setText(to_str(data.get("loc", np.nan)))
		self.scale.setText(to_str(data.get("scale", np.nan)))
		self.shape.setText(to_str(data.get("shape", np.nan)))
		self.minimum.setText(to_str(data.get("minimum", np.nan)))
		self.maximum.setText(to_str(data.get("maximum", np.nan)))
		self._check_negative()

	@property
	def _distribution_loc_label(self) -> str:
		if self.dist.id == sa.LognormalUncertainty.id:
			return "Loc (ln(mean)):"
		elif self.dist.id == sa.TriangularUncertainty.id:
			return "Mode:"
		elif self.dist.id == sa.BetaUncertainty.id:
			return "Loc / alpha:"
		elif self.dist.id in {sa.GammaUncertainty.id, sa.WeibullUncertainty.id}:
			return "Loc / offset:"
		else:
			return "Mean:"

	def _hide_params(self, *params, hide: bool = True) -> None:
		if "loc" in params:
			self.loc_label.setHidden(hide)
			self.loc.setHidden(hide)
		if "scale" in params:
			self.scale_label.setHidden(hide)
			self.scale.setHidden(hide)
		if "shape" in params:
			self.shape_label.setHidden(hide)
			self.shape.setHidden(hide)
		if "min" in params:
			self.min_label.setHidden(hide)
			self.minimum.setHidden(hide)
		if "max" in params:
			self.max_label.setHidden(hide)
			self.maximum.setHidden(hide)

	def _on_distribution_changed(self, index: int) -> None:
		self.dist = sa.uncertainty_choices[index]

		# Show/hide fields per distribution (mirror wizard)
		if self.dist.id in {0, 1}:  # Undefined / NoUncertainty
			self._hide_params("loc", "scale", "shape", "min", "max")
		elif self.dist.id in {2, 3}:  # Normal / Lognormal
			self._hide_params("shape", "min", "max")
			self._hide_params("loc", "scale", hide=False)
		elif self.dist.id in {4, 7}:  # Uniform / DiscreteUniform
			self._hide_params("loc", "scale", "shape")
			self._hide_params("min", "max", hide=False)
		elif self.dist.id in {5, 6}:  # Triangular / Bernoulli-like (min/max/loc)
			self._hide_params("scale", "shape")
			self._hide_params("loc", "min", "max", hide=False)
		elif self.dist.id in {8, 9, 10, 11, 12}:  # Other 3-param
			self._hide_params("min", "max")
			self._hide_params("loc", "scale", "shape", hide=False)

		# Special handling (lognormal and calculated mean label)
		if self.dist.id == sa.LognormalUncertainty.id:
			self.mean.setHidden(False)
			self.mean_label.setHidden(False)
			# Convert existing loc to log-space if coming from non-lognormal
			if self.previous_dist_id is not None and self.previous_dist_id != sa.LognormalUncertainty.id:
				self._extract_lognormal_loc_from_mean()
				self._sync_mean_from_loc()
		else:
			self.mean.setHidden(True)
			self.mean_label.setHidden(True)
			# If switching away from lognormal, set loc to linear amount if mean present
			if self.previous_dist_id == sa.LognormalUncertainty.id:
				try:
					mean_val = float(self.mean.text()) if self.mean.text() else np.nan
					if not np.isnan(mean_val):
						self.loc.setText(str(mean_val))
				except Exception:
					pass

		# Calculated mean visibility
		show_calc = self.dist.id in self.mean_is_calculated
		self.calc_mean_label.setHidden(not show_calc)
		self.calc_mean.setHidden(not show_calc)

		# Update labels
		self.loc_label.setText(self._distribution_loc_label)
		self.previous_dist_id = self.dist.id
		self.fields_box.updateGeometry()

		# Update plot and OK state
		self._generate_plot()
		self._update_ok_state()

	def _extract_lognormal_loc_from_mean(self) -> None:
		"""Set loc to ln(mean) when switching to lognormal, if mean is known."""
		try:
			mtxt = self.mean.text().strip()
			if not mtxt:
				return
			val = float(mtxt)
			if val == 0:
				self.loc.setText("nan")
			else:
				val = -1 * val if val < 0 else val
				self.loc.setText(str(np.log(val)))
		except Exception:
			self.loc.setText("nan")

	def _sync_mean_from_loc(self) -> None:
		if not self.loc.text():
			return
		try:
			self.mean.setText(str(np.exp(float(self.loc.text()))))
		except Exception:
			self.mean.setText("nan")
		self._update_ok_state()

	def _sync_loc_from_mean(self) -> None:
		if not self.mean.hasAcceptableInput():
			self.loc.setText("nan")
			self._update_ok_state()
			return
		try:
			val = float(self.mean.text()) if self.mean.text() else float("nan")
		except Exception:
			val = float("nan")
		if np.isnan(val) or val == 0:
			self.loc.setText("nan")
		else:
			val = -1 * val if val < 0 else val
			self.loc.setText(str(np.log(val)))
		self._update_ok_state()

	def _check_negative(self) -> None:
		# Special case for lognormal negative mean
		try:
			if not self.mean.hasAcceptableInput():
				return
			val = float(self.mean.text()) if self.mean.text() else float("nan")
		except Exception:
			val = float("nan")
		self.negative.setChecked(bool(not np.isnan(val) and val < 0))

	def _standard_dist_fields(self, dist_id: int) -> list:
		if dist_id in {2, 3}:
			return ["loc", "scale"]
		elif dist_id in {4, 7}:
			return ["minimum", "maximum"]
		elif dist_id in {5, 6}:
			return ["loc", "minimum", "maximum"]
		elif dist_id in {8, 9, 10, 11, 12}:
			return ["loc", "scale", "shape"]
		else:
			return []

	@property
	def _uncertainty_info(self) -> dict:
		data = {k: v for k, v in EMPTY_UNCERTAINTY.items()}
		data["uncertainty type"] = self.distribution.currentIndex()
		data["negative"] = bool(self.negative.isChecked())
		# Pull values from widgets
		def as_float(txt: str) -> float:
			try:
				val = float(txt)
				return val
			except Exception:
				return float("nan")

		for field in self._standard_dist_fields(data["uncertainty type"]):
			widget = {
				"loc": self.loc,
				"scale": self.scale,
				"shape": self.shape,
				"minimum": self.minimum,
				"maximum": self.maximum,
			}[field]
			data[field] = as_float(widget.text())
		return data

	def _completed_active_fields(self) -> bool:
		# Mirror wizard validations
		dist_id = self.dist.id
		def ok_lineedit(le: QtWidgets.QLineEdit) -> bool:
			return bool(le.hasAcceptableInput() and le.text())

		if dist_id in {0, 1}:
			return True
		elif dist_id in {2, 3}:
			return ok_lineedit(self.loc) and ok_lineedit(self.scale)
		elif dist_id in {4, 7}:
			return ok_lineedit(self.minimum) and ok_lineedit(self.maximum)
		elif dist_id in {5, 6}:
			if not (ok_lineedit(self.minimum) and ok_lineedit(self.maximum) and ok_lineedit(self.loc)):
				return False
			try:
				return float(self.minimum.text()) < float(self.loc.text()) < float(self.maximum.text())
			except Exception:
				return False
		elif dist_id in {8, 9, 10, 11, 12}:
			return ok_lineedit(self.scale) and ok_lineedit(self.shape) and ok_lineedit(self.loc)
		return False

	def _update_ok_state(self) -> None:
		ok_btn = self.buttons.button(QtWidgets.QDialogButtonBox.Ok)
		ok_btn.setEnabled(self._completed_active_fields())

	def _generate_plot(self) -> None:
		# Update calculated mean if applicable and render sample
		if self.dist is None:
			return
		complete = self._completed_active_fields() or self.dist.id in {sa.UndefinedUncertainty.id, sa.NoUncertainty.id}
		if not complete:
			self._update_ok_state()
			return
		array = self.dist.from_dicts(self._uncertainty_info)
		# Calculated mean display for specific distributions
		if self.dist.id in self.mean_is_calculated:
			try:
				calc = self.dist.statistics(array).get("mean")
			except TypeError:
				# DiscreteUniform workaround
				array = self.dist.fix_nan_minimum(array)
				calc = (array["maximum"] + array["minimum"]) / 2
			calc = calc.mean() if isinstance(calc, np.ndarray) else calc
			self.calc_mean.setText(str(float(calc)))
		# Vertical line value
		if self.dist.id == sa.LognormalUncertainty.id:
			vline = self.dist.statistics(array).get("median")
		elif self.dist.id in {sa.UndefinedUncertainty.id, sa.NoUncertainty.id}:
			# Best effort: use loc as "mean" placeholder
			try:
				vline = float(self.loc.text()) if self.loc.text() else np.nan
			except Exception:
				vline = np.nan
		else:
			vline = self.dist.statistics(array).get("mean")
		# Sample data
		data = self.dist.random_variables(array, 1000)
		if not np.any(np.isnan(data)):
			try:
				self.plot.plot(data, vline)
			except RuntimeError as e:
				logger.error("%s: plotting failed, retry without KDE", e)
				try:
					sns.histplot(data.T, kde=False, stat="density", ax=self.plot.ax, edgecolor="none")
					self.plot.ax.axvline(vline, label="Mean / amount", c="r", ymax=0.98)
					self.plot.ax.legend(loc="upper right")
					self.plot.canvas.draw()
				except Exception:
					pass
		self._update_ok_state()

	def _on_accept(self) -> None:
		try:
			self.result_dict = self._uncertainty_info
			self.result_array = self.dist.from_dicts(self._uncertainty_info)
		except Exception as e:
			QtWidgets.QMessageBox.warning(
				self,
				"Invalid uncertainty",
				str(e),
				QtWidgets.QMessageBox.Ok,
				QtWidgets.QMessageBox.Ok,
			)
			return
		self.accept()


class SimpleDistributionPlot(ABPlot):
	def plot(self, data: np.ndarray, mean: float, label: str = "Value"):
		self.reset_plot()
		try:
			sns.histplot(data.T, kde=True, stat="density", ax=self.ax, edgecolor="none")
		except RuntimeError as e:
			logger.error("%s: Plotting without KDE.", e)
			sns.histplot(data.T, kde=False, stat="density", ax=self.ax, edgecolor="none")
		self.ax.set_xlabel(label)
		self.ax.set_ylabel("Probability density")
		# Add vertical line at given mean of x-axis
		self.ax.axvline(mean, label="Mean / amount", c="r", ymax=0.98)
		self.ax.legend(loc="upper right")
		_, height = self.canvas.get_width_height()
		self.setMinimumHeight(height / 2)
		self.canvas.draw()


__all__ = ["UncertaintyDialog"]




from typing import Any

import bw2data as bd

from activity_browser.app.actions.base import ABAction, exception_dialogs
from activity_browser import app
from activity_browser.ui.dialogs import UncertaintyDialog
from activity_browser.ui.icons import qicons


class ParameterUncertaintyModify(ABAction):
    """
    ABAction to modify the uncertainty of an existing parameter.
    """

    icon = qicons.edit
    text = "Modify parameter uncertainty"

    @staticmethod
    @exception_dialogs
    def run(parameter: Any, uncertainty_dict: dict=None) -> None:

        if not uncertainty_dict:
            initial = parameter.dict.copy() if "uncertainty type" in parameter.dict else None

            ok, uncertainty_dict = UncertaintyDialog.get_uncertainty_dict(
                parent=app.main_window,
                initial=initial,
                )
        
            if not ok:
                return

        parameter.data.update(uncertainty_dict)
        parameter.save()
        bd.parameters.recalculate()



from loguru import logger

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Qt

import pandas as pd
import bw2data as bd
from bw2data.parameters import ParameterizedExchange
from bw2data.backends import ExchangeDataset

from activity_browser import app
from activity_browser.ui import widgets, icons, delegates, core
from activity_browser.bwutils.commontasks import database_is_locked
from activity_browser.bwutils.utils import Parameter


class ParameterizedExchangesSection(QtWidgets.QWidget):
    """
    A widget section that displays all parameterized exchanges in the current project.

    Attributes:
        model (ParameterizedExchangesModel): The model containing the data for the exchanges.
        view (ParameterizedExchangesView): The view displaying the exchanges.
    """

    def __init__(self, parent=None):
        """
        Initializes the ParameterizedExchangesSection widget.

        Args:
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self._populate_later_flag = False

        # Parameterized exchanges table view
        self.model = ParameterizedExchangesModel(parent=self)
        self.view = ParameterizedExchangesView()
        self.view.setModel(self.model)

        self.build_layout()
        self.connect_signals()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.view)
        self.setLayout(layout)

    def connect_signals(self):
        """
        Connects signals to their respective slots.
        """
        app.signals.metadata.synced.connect(self.syncLater)
        app.signals.parameter.changed.connect(self.syncLater)
        app.signals.parameter.recalculated.connect(self.syncLater)
        app.signals.parameter.deleted.connect(self.syncLater)
        app.signals.project.changed.connect(self.syncLater)
        app.signals.meta.databases_changed.connect(self.syncLater)

    def syncLater(self):
        """
        Schedules a sync operation to be performed later.
        """

        def slot():
            self._populate_later_flag = False
            self.sync()
            self.thread().eventDispatcher().awake.disconnect(slot)

        if self._populate_later_flag:
            return

        self._populate_later_flag = True
        self.thread().eventDispatcher().awake.connect(slot)

    def sync(self):
        """
        Synchronizes the widget with the current state of parameterized exchanges.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        df = self.build_exchanges_df()
        self.model.set_dataframe(df)

    def build_exchanges_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from all parameterized exchanges in the project.

        Returns:
            pd.DataFrame: The DataFrame containing the parameterized exchanges data.
        """
        translated = []

        # Get all parameterized exchanges
        for param_exc in ParameterizedExchange.select():
            try:
                exchange = bd.Edge(document=ExchangeDataset.get_by_id(param_exc.exchange))

                # Get keys for input and output
                input_key = exchange.get("input")
                output_key = exchange.get("output")

                # Get metadata from metadata store
                input_meta = app.metadata.get_metadata([input_key], ["name", "unit", "location", "database", "product"]).iloc[0]
                output_meta = app.metadata.get_metadata([output_key], ["name"]).iloc[0]

                row = {
                    "amount": exchange.get("amount"),
                    "unit": input_meta.get("unit"),
                    "from": input_meta.get("product") or input_meta.get("name"),
                    "to": output_meta.get("name"),
                    "database": input_meta.get("database"),
                    "formula": exchange.get("formula"),
                    "comment": exchange.get("comment"),
                    "uncertainty": exchange.get("uncertainty type"),
                    "_exchange": exchange,
                    "_output_key": output_key,
                    "_input_key": input_key,
                }
                translated.append(row)
            except Exception as e:
                # Skip if exchange can't be loaded
                continue

        columns = ["amount", "unit", "from", "to", "database", "formula", "comment", "uncertainty", "_exchange", "_output_key", "_input_key"]
        return pd.DataFrame(translated, columns=columns)


class ParameterizedExchangesView(widgets.ABTreeView):
    """
    A view that displays parameterized exchanges in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "amount": delegates.FloatDelegate,
        "unit": delegates.StringDelegate,
        "product": delegates.StringDelegate,
        "producer": delegates.StringDelegate,
        "location": delegates.StringDelegate,
        "database": delegates.StringDelegate,
        "formula": delegates.NewFormulaDelegate,
        "comment": delegates.StringDelegate,
        "uncertainty": delegates.UncertaintyDelegate,
    }

    class ContextMenu(QtWidgets.QMenu):
        """
        A context menu for the ParameterizedExchangesView.
        """
        def __init__(self, pos, view: "ParameterizedExchangesView"):
            """
            Initializes the ContextMenu.

            Args:
                pos: The position of the context menu.
                view (ParameterizedExchangesView): The view displaying the exchanges.
            """
            super().__init__(view)

            index = view.indexAt(pos)
            if index.isValid() and not view.model().isBranchNode(index):
                row = view.model().row(index)
                if row is not None:
                    output_key = row.get("_output_key")
                    if output_key:
                        # Open activity action
                        open_action = app.actions.ActivityOpen.get_QAction([output_key])
                        open_action.setText("Open activity")
                        self.addAction(open_action)


class ParameterizedExchangesModel(core.ABTreeModel):
    """
    A model representing the data for parameterized exchanges.
    """

    def __init__(self, parent=None):
        """
        Initializes the ParameterizedExchangesModel.

        Args:
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(df=pd.DataFrame(), parent=parent)

    def setData(self, index: QtCore.QModelIndex, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        """
        Sets the data for the given index.

        Args:
            index (QtCore.QModelIndex): The index to set data for.
            value: The value to set.
            role (int): The role for which to set the data.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if role != Qt.ItemDataRole.EditRole:
            return False

        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return False

        exchange = row.get("_exchange")
        if exchange is None:
            return False

        if column_name in ["amount", "formula", "comment"]:
            if column_name == "formula" and not str(value).strip():
                # Remove formula if empty
                app.actions.ExchangeFormulaRemove.run([exchange])
                return True

            app.actions.ExchangeModify.run(exchange, {column_name.lower(): value})
            return True

        return False

    def decorationData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides decoration data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide decoration data.

        Returns:
            The decoration data for the index.
        """
        column_name = self.column_name(index)

        if column_name == "amount":
            formula = self.get(index, "formula")
            if pd.isna(formula) or formula is None or formula == "":
                return icons.qicons.edit
            return icons.qicons.parameterized

        return None

    def indexEditable(self, index: QtCore.QModelIndex) -> bool:
        """
        Returns whether the index is editable.

        Args:
            index (QtCore.QModelIndex): The index to check.

        Returns:
            bool: True if the index is editable, False otherwise.
        """
        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return False

        # Check if database is locked
        exchange = row.get("_exchange")
        if exchange and database_is_locked(exchange.output["database"]):
            return False

        # Allow editing for specific columns
        if column_name in ["amount", "formula", "comment"]:
            return True

        return False

    def scoped_parameters(self, index: QtCore.QModelIndex) -> dict[str, Parameter]:
        """
        Returns the parameters in scope of the exchange at the given index.

        Args:
            index (QtCore.QModelIndex): The index to get scoped parameters for.

        Returns:
            dict: The parameters in scope.
        """
        from activity_browser.bwutils.commontasks import parameters_in_scope

        row = self.row(index)
        if row is None:
            return {}

        exchange = row.get("_exchange")
        if exchange is None:
            return {}

        return parameters_in_scope(node=exchange.output)




from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt
from loguru import logger

import pandas as pd
import bw2data as bd

from bw2data.parameters import ProjectParameter, DatabaseParameter, ActivityParameter, Group, ParameterBase

from activity_browser import app
from activity_browser.ui import widgets, icons, delegates, core
from activity_browser.bwutils.commontasks import refresh_node, refresh_parameter, parameters_in_scope, database_is_locked, node_group
from activity_browser.bwutils.utils import Parameter


class ParametersTab(QtWidgets.QWidget):
    """
    A widget that displays parameters related to a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display parameters for.
        model (ParametersModel): The model containing the data for the parameters.
        view (ParametersView): The view displaying the parameters.
    """
    def __init__(self, activity, parent=None):
        """
        Initializes the ParametersTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display parameters for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.activity = refresh_node(activity)

        self.view = ParametersView(self)
        self.view.setSortingEnabled(False)
        self.view.setUniformRowHeights(True)

        self.model = ParametersModel(tab=self)
        self.view.setModel(self.model)

        self.build_layout()
        self.connect_signals()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 10, 0, 1)
        layout.addWidget(self.view)

        self.setLayout(layout)

    def connect_signals(self):
        """
        Connects signals to their respective slots.
        """
        app.signals.parameter.changed.connect(self.sync)
        app.signals.parameter.recalculated.connect(self.sync)
        app.signals.parameter.deleted.connect(self.sync)

    def sync(self):
        """
        Synchronizes the widget with the current state of the activity.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        df = self.build_df()
        self.model.set_dataframe(df, group=["_param_type", "_scope"])
        self.view.expandAll()

        self.view.resizeColumnToContents(1)
        self.view.resizeColumnToContents(3)
        self.view.resizeColumnToContents(4)

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from all parameters in the project.

        Returns:
            pd.DataFrame: The DataFrame containing the parameters data.
        """
        translated = []

        # Project parameters
        for param in ProjectParameter.select():
            row = self._parameter_to_row(param)
            translated.append(row)

        translated.append({
            "name": "New parameter...",
            "_group": "project",
            "_param_type": "project",
            "_class": "new",
        })

        # Database parameters
        db_params = DatabaseParameter.select()
        db_name = self.activity["database"]

        for param in db_params.where(DatabaseParameter.database == db_name):
            row = self._parameter_to_row(param, db_name, db_name)
            translated.append(row)

        if not database_is_locked(db_name):
            translated.append({
                "name": "New parameter...",
                "_scope": db_name,
                "_database": db_name,
                "_group": db_name,
                "_param_type": "database",
                "_class": "new",
            })

        # Activity parameters
        act_params = ActivityParameter.select()
        group_name = node_group(self.activity) or str(self.activity.id)

        for param in act_params.where(ActivityParameter.group == group_name):
            row = self._parameter_to_row(param, f"Group: {group_name}", param.database)
            translated.append(row)

        if not database_is_locked(self.activity["database"]):
            translated.append({
                "name": "New parameter...",
                "_scope": f"Group: {group_name}",
                "_database": self.activity["database"],
                "_group": group_name,
                "_param_type": "activity",
                "_class": "new",
            })

        columns = ["name", "amount", "formula", "uncertainty", "comment", "_parameter", "_scope", "_database", "_group",
                   "_param_type", "_class"]
        df = pd.DataFrame(translated, columns=columns)

        df["_activity"] = [self.activity for i in range(len(df))]
        return df

    def _parameter_to_row(self, param, scope_label: str = None, database: str = None) -> dict:
        """
        Converts a parameter to a row dictionary.

        Args:
            param: The parameter to convert (ProjectParameter, DatabaseParameter, or ActivityParameter).
            scope_label: The label for the scope (e.g., "Current project", "Database: ecoinvent").
            database: The database name (None for project parameters).

        Returns:
            dict: A dictionary representing the parameter row.
        """
        data = param.dict

        # Create Parameter wrapper
        if isinstance(param, ProjectParameter):
            parameter = Parameter(param.name, "project", data.get("amount"), data, "project")
            group = "project"
            param_type = "project"
        elif isinstance(param, DatabaseParameter):
            parameter = Parameter(param.name, param.database, data.get("amount"), data, "database")
            group = param.database
            param_type = "database"
        elif isinstance(param, ActivityParameter):
            parameter = Parameter(param.name, param.group, data.get("amount"), data, "activity")
            group = param.group
            param_type = "activity"
        else:
            raise ValueError(f"Unknown parameter type: {type(param)}")

        row = {
            "name": parameter.name,
            "amount": parameter.amount,
            "uncertainty": parameter.uncertainty,
            "formula": data.get("formula"),
            "comment": data.get("comment"),
            "_param_type": param_type,
            "_parameter": parameter,
            "_scope": scope_label,
            "_database": database,
            "_group": group,
            "_class": "instantiated",
        }

        return row


class ParametersView(widgets.ABTreeView):
    """
    A view that displays the parameters in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "amount": delegates.FloatDelegate,
        "name": delegates.StringDelegate,
        "formula": delegates.NewFormulaDelegate,
        "comment": delegates.StringDelegate,
        "uncertainty": delegates.UncertaintyDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m: m.add(app.actions.ParameterDelete, m.parameters, enable=bool(m.parameters) and not m.locked),
        ]

        @property
        def locked(self):
            table_view: ParametersView = self.parent()
            return database_is_locked(table_view.activity["database"])
        
        @property
        def activity(self):
            table_view: ParametersView = self.parent()
            return table_view.activity
        
        @property
        def parameters(self):
            table_view: ParametersView = self.parent()
            table_model: ParametersModel = table_view.model()
            
            selected_indices = table_view.selectedIndexes()
            params = table_model.values_from_indices("_parameter", selected_indices)
            # Convert to peewee models
            return [p.to_peewee_model() for p in params if p is not None]
    
    @property
    def activity(self):
        """
        Returns the activity associated with the view.

        Returns:
            The activity associated with the view.
        """
        return self.parent().activity


class ParametersModel(core.ABTreeModel):
    """
    A model representing the data for the parameters.
    """
    def __init__(self, tab: ParametersTab):
        super().__init__(parent=tab)
        self.tab = tab
    
    def setData(self, index: QtCore.QModelIndex, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        """
        Sets the data for the given index.

        Args:
            index (QtCore.QModelIndex): The index to set data for.
            value: The value to set.
            role (int): The role for which to set the data.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if role != Qt.ItemDataRole.EditRole:
            return False

        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return False

        # Handle "New parameter..." rows
        if row.get("_class") == "new":
            if column_name != "name" or value == "":
                return False

            parameter = Parameter(
                name=value,
                group=row.get("_group"),
                param_type=row.get("_param_type")
            )

            app.actions.ParameterNewFromParameter.run(parameter)
            return True

        # Handle regular parameter edits
        parameter = row.get("_parameter")
        if parameter is None:
            return False

        if column_name in ["amount", "formula", "name", "comment"]:
            parameter = refresh_parameter(parameter)
            app.actions.ParameterModify.run(parameter, column_name, value)

        if column_name == "uncertainty":
            parameter = refresh_parameter(parameter)
            app.actions.ParameterUncertaintyModify.run(parameter.to_peewee_model(), uncertainty_dict=value)

            return True

        return False

    def decorationData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides decoration data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide decoration data.

        Returns:
            The decoration data for the index.
        """
        column_name = self.column_name(index)

        if column_name == "amount":
            formula = self.get(index, "formula")
            formula = isinstance(formula, str) and formula.strip()

            return icons.qicons.parameterized if formula else icons.qicons.empty

        return None

    def fontData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides font data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide font data.

        Returns:
            QtGui.QFont: The font data for the index.
        """
        param_class = self.get(index, "_class")
        if param_class == "new":
            font = QtGui.QFont()
            font.setWeight(QtGui.QFont.Weight.ExtraLight)
            return font

        if param_class == "broken":
            font = QtGui.QFont()
            font.setWeight(QtGui.QFont.Weight.Bold)
            return font

        return None

    def indexEditable(self, index: QtCore.QModelIndex) -> bool:
        """
        Returns whether the index is editable.

        Args:
            index (QtCore.QModelIndex): The index to check.

        Returns:
            bool: True if the index is editable, False otherwise.
        """
        column_name = self.column_name(index)

        # Check if database is locked
        database = self.get(index, "_database")
        if not pd.isna(database) and database_is_locked(database):
            return False

        # Prevent editing broken parameters
        if self.get(index, "_class") == "broken":
            return False

        # Allow editing for specific columns
        if column_name in ["formula", "uncertainty", "name", "comment"]:
            return True

        if column_name == "amount" and not self.get(index, "formula"):
            return True

        return False

    def scoped_parameters(self, index: QtCore.QModelIndex) -> dict[str, Parameter]:
        """
        Returns the parameters in scope of the parameter at the given index.

        Args:
            index (QtCore.QModelIndex): The index to get scoped parameters for.

        Returns:
            dict: The parameters in scope.
        """
        parameter = self.get(index, "_parameter")
        if parameter is None or isinstance(parameter, float):  # NaN check
            return {}

        return parameters_in_scope(parameter=parameter)



from qtpy import QtWidgets, QtCore
from loguru import logger

import pandas as pd
import bw2data as bd
import bw_functional as bf

from activity_browser import app
from activity_browser.bwutils.commontasks import refresh_node, database_is_locked
from activity_browser.ui import widgets, delegates, core


class DataTab(QtWidgets.QWidget):
    """
    A widget that displays the data structure of a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display data for.
        data_view (DataView): The view displaying the data.
        data_model (DataModel): The model containing the data.
    """
    def __init__(self, activity: tuple | int | bd.Node, parent=None):
        """
        Initializes the DataTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display data for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)

        self.activity = refresh_node(activity)

        # Data TreeView
        self.data_view = DataView(self)
        self.data_model = DataModel(parent=self)
        self.data_view.setModel(self.data_model)

        df = self.build_df()
        df.reset_index(drop=True, inplace=True)
        self.data_model.set_dataframe(df)
        self.data_model.group(["_name"])
        self.data_view.expandAll()

        self.build_layout()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.data_view)
        self.setLayout(layout)

    def sync(self) -> None:
        """
        Synchronizes the widget with the current state of the activity.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        self.activity = refresh_node(self.activity)
        df = self.build_df()
        df.reset_index(drop=True, inplace=True)
        self.data_model.set_dataframe(df)
        self.data_model.group(["_name"])
        self.data_view.expandAll()

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the activity data.

        Returns:
            pd.DataFrame: The DataFrame containing the activity data.
        """
        df = pd.Series(self.activity.as_dict()).to_frame()
        df["_name"] = f"{self.activity['name']} {df.get('product', '')} ({self.activity['id']})"
        df["_activity_id"] = self.activity.id
        df["_activity_db"] = self.activity["database"]

        if isinstance(self.activity, bf.Process):
            for product in self.activity.products():
                fn_df = pd.DataFrame.from_dict(product.as_dict(), orient="index")
                fn_df["_name"] = f"{product['name']}: {product.get('product', '')} ({product['id']})"
                fn_df["_activity_id"] = product.id
                fn_df["_activity_db"] = product["database"]
                df = pd.concat([df, fn_df])

        df = df.reset_index()
        df = df.rename({"index": "field", 0: "value"}, axis=1)
        df = df.sort_values(["_name", "field"], ignore_index=True)

        cols = ["field", "value", "_name", "_activity_id", "_activity_db"]
        return df[cols]


class DataView(widgets.ABTreeView):
    """
    A view that displays the data in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "field": delegates.StringDelegate,
        "value": delegates.NewFormulaDelegate,
    }


class DataModel(core.ABTreeModel):
    """
    A model representing the data for the activity.
    """
    
    def setData(self, index: QtCore.QModelIndex, value, role: int = QtCore.Qt.ItemDataRole.EditRole) -> bool:
        """
        Sets the data for the given index.

        Args:
            index (QtCore.QModelIndex): The index to set data for.
            value: The value to set.
            role (int): The role for which to set the data.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if role != QtCore.Qt.ItemDataRole.EditRole:
            return False

        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return False

        if column_name == "value":
            value = eval(value)
            app.actions.ActivityModify.run(row.get("_activity_id"), row.get("field"), value)
            return True

        return False
    
    def indexEditable(self, index: QtCore.QModelIndex) -> bool:
        """
        Returns whether the index is editable.

        Args:
            index (QtCore.QModelIndex): The index to check.

        Returns:
            bool: True if the index is editable, False otherwise.
        """
        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return False

        if column_name == "value" and not database_is_locked(row.get("_activity_db")):
            return True
        
        return False
    
    def displayData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides display data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide display data.

        Returns:
            The display data for the index.
        """
        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            # Branch node
            node = index.internalPointer()
            if isinstance(node, core.TreeNode):
                return node.path[-1] if index.column() == 0 else None
            return None

        if column_name == "value":
            data = row.get(column_name)
            if isinstance(data, str):
                return f"'{data}'"
            return str(data)

        return row.get(column_name)



import numbers
import json
from datetime import datetime as dt
from pathlib import Path
from typing import Union

import xlsxwriter
import bw2data as bd
from bw2io.export.csv import reformat
from bw2io.export.excel import CSVFormatter, create_valid_worksheet_name

from .importers import ABPackage
from .pedigree import PedigreeMatrix

# Copied most of this code wholesale from bw2io package.
# TODO: reminder to make a pull-request for these things in bw2io repo.
#  - Add the 'nan_inf_to_errors' option when opening the xlsxwriter.Workbook.
#  - Add handler for pedigree data to exporter
#  - Add 'database' field as required CSVFormatter export field.
#  - Add code to ensure no second 'activity' field is exported, as this
#  messes with following import.


def ab_reformat(value):
    if isinstance(value, dict):
        try:
            return json.dumps(value)
        except TypeError:
            return "Non-serializable dictionary"
    return reformat(value)


class ABCSVFormatter(CSVFormatter):
    def get_activity_metadata(self, act):
        excluded = {"database", "name", "activity"}
        return {
            "name": act.get("name"),
            "metadata": sorted(
                [
                    (k, ab_reformat(v))
                    for k, v in act.items()
                    if k not in excluded and not isinstance(v, list)
                ]
            ),
            "parameters": self.get_activity_parameters(act),
        }

    def exchange_as_dict(self, exc):
        """Same as CSVFormatter, but explicitly pull the database from the
        input activity.

        This ensures that the database value is always included, even when
        it is not stored in the exchange _data.
        """
        inp = exc.input
        inp_fields = (
            "name",
            "unit",
            "location",
            "categories",
            "database",
            "reference product",
        )
        skip_fields = ("input", "output")
        data = {k: v for k, v in exc._data.items() if k not in skip_fields}
        if "product" in data and "reference product" not in data:
            data["reference product"] = data.pop("product")
        data.update(**{k: inp[k] for k in inp_fields if inp.get(k)})
        return data


def format_pedigree(data: dict) -> str:
    """Converts pedigree dict to tuple."""
    try:
        return "::".join(
            str(x) for x in PedigreeMatrix.from_dict(data).factors_as_tuple()
        )
    except AssertionError:
        # Will skip any other kinds of dictionaries.
        return "nan"


def frmt_str(data: Union[str, dict]) -> str:
    """Format non-numerical data (like tuples) to string format."""
    return format_pedigree(data) if isinstance(data, dict) else str(data)


def write_lci_excel(db_name: str, path: str, objs=None, sections=None) -> Path:
    """Export database `database_name` to an Excel spreadsheet.

    Not all data can be exported. The following constraints apply:

    * Nested data, e.g. `{'foo': {'bar': 'baz'}}` are excluded. Spreadsheets are not a great format for nested data.
      However, *tuples* are exported, and the characters `::` are used to join elements of the tuple.
    * The only well-supported data types are strings, numbers, and booleans.

    Returns the filepath of the exported file.

    """
    path = Path(path)
    if not path.suffix == ".xlsx":
        out_file = path / "lci-{}.xlsx".format(bd.utils.safe_filename(db_name, False))
    else:
        out_file = path

    workbook = xlsxwriter.Workbook(out_file, {"nan_inf_to_errors": True})
    bold = workbook.add_format({"bold": True})
    bold.set_font_size(12)
    highlighted = {
        "Activity",
        "Database",
        "Exchanges",
        "Parameters",
        "Database parameters",
        "Project parameters",
    }
    frmt = lambda x: bold if row[0] in highlighted else None

    sheet = workbook.add_worksheet(create_valid_worksheet_name(db_name))

    data = ABCSVFormatter(db_name, objs).get_formatted_data(sections)

    for row_index, row in enumerate(data):
        for col_index, value in enumerate(row):
            if value is None:
                continue
            elif isinstance(value, numbers.Number):
                sheet.write_number(row_index, col_index, value, frmt(value))
            else:
                sheet.write_string(row_index, col_index, frmt_str(value), frmt(value))

    workbook.close()

    return out_file


def store_database_as_package(db_name: str, directory: str = None) -> bool:
    """Attempt to use `bw.BW2Package` to save the given database as an
    isolated package that can be shared with others.
    Returns a boolean signifying success or failure.
    """
    if db_name not in bd.databases:
        return False
    metadata = bd.databases[db_name]
    db = bd.Database(db_name)
    directory = directory or bd.projects.output_dir
    output_dir = Path(directory)
    if output_dir.suffix == ".bw2package":
        out_file = output_dir
    else:
        out_file = output_dir / "{}.bw2package".format(db.filename)
    # First, ensure the metadata on the database is up-to-date.
    modified = dt.strptime(metadata["modified"], "%Y-%m-%dT%H:%M:%S.%f")
    if "processed" in metadata:
        processed = dt.strptime(metadata["processed"], "%Y-%m-%dT%H:%M:%S.%f")
        if processed < modified:
            db.process()
    else:
        db.process()
    # Now that processing is done, perform the export.
    ABPackage.unrestricted_export(db, out_file)
    return True



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets


class BaseSettingsChapter(QtWidgets.QWidget):
    """Base class for settings chapters."""
    
    # Signal emitted when settings change
    changed = QtCore.Signal()
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.settings_page = parent
        self._initial_state = None
    
    def get_current_state(self):
        """
        Override this to return the current state of the chapter.
        Should return a dictionary or tuple representing current values.
        """
        return {}
    
    def has_changes(self):
        """Check if the chapter has unsaved changes."""
        if self._initial_state is None:
            return False
        return self.get_current_state() != self._initial_state
    
    def save_settings(self):
        """Override this to save chapter-specific settings."""
        pass
    
    def reset(self):
        """Override this to reset chapter to initial values."""
        pass
    
    def restore_defaults(self):
        """Override this to restore default values."""
        pass



from qtpy import QtWidgets, QtCore, QtGui

import pandas as pd

from bw2io.importers.base_lci import LCIImporter

from activity_browser.ui import widgets, core

from .node_tab import ImportPreviewNodeTab
from .edge_tab import ImportPreviewEdgeTab


class ImportPreviewDialog(QtWidgets.QDialog):
    def __init__(self, importer: LCIImporter, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Import Preview")
        self.resize(600, 400)

        self.importer = importer
        self.tabs = QtWidgets.QTabWidget(self)

        self.node_tab = ImportPreviewNodeTab(importer, self)
        self.edge_tab = ImportPreviewEdgeTab(importer, self)

        self.tabs.addTab(self.node_tab, "Nodes")
        self.tabs.addTab(self.edge_tab, "Edges")

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.tabs)
        self.setLayout(layout)



import json
import os
from loguru import logger

from qtpy import QtWebChannel, QtWebEngineWidgets, QtWidgets
from qtpy.QtCore import QObject, Qt, QUrl, Signal, SignalInstance, Slot

import bw2data as bd
import bw_functional as bf

from activity_browser import static, app
from activity_browser.bwutils.commontasks import refresh_node, database_is_locked
from activity_browser.ui import widgets
from .exchanges_tab import get_exchange_type




class GraphTab(QtWidgets.QWidget):
    """
    A widget that displays a graph related to a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display the graph for.
        expanded_nodes (set): A set of node IDs that are expanded in the graph.
        button (QtWidgets.QPushButton): A button to trigger synchronization.
        bridge (Bridge): A bridge object for communication between Python and JavaScript.
        backend (GraphBackend): A backend object for communication between Python and JavaScript.
        url (QUrl): The URL of the HTML file to display.
        channel (QtWebChannel.QWebChannel): A web channel for communication between Python and JavaScript.
        page (Page): A web engine page to display the HTML content.
        view (QtWebEngineWidgets.QWebEngineView): A web engine view to display the HTML content.
    """
    def __init__(self, activity, parent=None):
        """
        Initializes the GraphTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display the graph for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.setAcceptDrops(True)

        self.activity = refresh_node(activity)
        self.expanded_nodes = {self.activity.id}

        self.button = QtWidgets.QPushButton("CLICK ME")
        self.button.clicked.connect(self.sync)

        self.bridge = Bridge(self)
        self.backend = GraphBackend(self)
        self.url = QUrl.fromLocalFile(os.path.join(static.__path__[0], "activity_graph.html"))

        self.channel = QtWebChannel.QWebChannel(self)
        self.channel.registerObject("bridge", self.bridge)
        self.channel.registerObject("backend", self.backend)

        self.page = Page()
        self.page.setWebChannel(self.channel)

        self.view = GraphView(self)
        self.view.setPage(self.page)
        self.view.setUrl(self.url)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        self.setLayout(layout)

        self.bridge.ready.connect(self.sync)

    def sync(self):
        """
        Synchronizes the widget with the current state of the activity.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        self.activity = refresh_node(self.activity)
        json = self.build_json()
        self.bridge.update_graph.emit(json)

    def build_json(self):
        """
        Builds a JSON representation of the graph.

        Returns:
            str: The JSON representation of the graph.
        """
        nodes = []
        edges = []

        collapsed_functions = set()
        for node_id in self.expanded_nodes:
            node = bd.get_node(id=node_id)
            excs = list(node.exchanges())
            function_nodes = [exc.input for exc in excs if exc["type"] == "production"]
            functions = []

            for fn_node in function_nodes:
                functions.append({
                    "id": f"bw{fn_node.id}",
                    "name": fn_node._document.product if fn_node._document.product else fn_node["name"]
                })
                excs.extend(fn_node.upstream())

            nodes.append({
                "id": f"bw{node.id}",
                "name": node["name"],
                "functions": functions,
                "type": "expanded_node"
            })

            for exc in excs:
                if exc["type"] in ["production", "biosphere"]:
                    continue
                processor = get_processor_from_exchange(exc)

                source_id = processor.id
                target_id = exc.output.id

                if source_id not in self.expanded_nodes:
                    source_id = exc.input.id
                    collapsed_functions.add(source_id)

                if target_id not in self.expanded_nodes:
                    collapsed_functions.add(target_id)

                edges.append({
                    "source_id": f"bw{source_id}",
                    "target_id": f"bw{exc.output.id}",
                    "function_id": f"bw{exc.input.id}",
                })

        for node_id in collapsed_functions:
            fn_node = bd.get_node(id=node_id)
            nodes.append({
                "id": f"bw{node_id}",
                "name": fn_node._document.product if fn_node._document.product else fn_node["name"],
                "functions": [],
                "type": "collapsed_function"
            })

        full = {
            "nodes": nodes,
            "edges": edges,
        }

        return json.dumps(full)

    def expand_node(self, node_id: str):
        """
        Expands a node in the graph.

        Args:
            node_id (str): The ID of the node to expand.
        """
        node_id = int(node_id)  # JS shenanigans can't deal with 64 bit strings
        node = bd.get_node(id=node_id)
        if isinstance(node, bf.Product):
            node = bd.get_node(key=node["processor"])
        self.expanded_nodes.add(node.id)
        self.sync()

    def collapse_node(self, node_id: str):
        """
        Collapses a node in the graph.

        Args:
            node_id (str): The ID of the node to collapse.
        """
        node_id = int(node_id)  # JS shenanigans can't deal with 64 bit strings
        if self.activity.id == node_id:
            return
        self.expanded_nodes.remove(int(node_id))
        self.sync()


def get_processor_from_exchange(exchange):
    """
    Gets the processor from an exchange.

    Args:
        exchange: The exchange to get the processor from.

    Returns:
        The processor of the exchange.
    """
    source = exchange.input
    processors = list(source.upstream(kinds=["production"]))
    if len(processors) > 1:
        logger.warning("Multiple processors, only taking first one")
    processor = processors[0]
    return processor.output


class GraphView(QtWebEngineWidgets.QWebEngineView):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.setContextMenuPolicy(Qt.PreventContextMenu)
        self.overlay = None

    def dragEnterEvent(self, event):
        """
        Handles the drag enter event.

        Args:
            event: The drag enter event.
        """
        if database_is_locked(self.parent().activity["database"]):
            return

        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            self.overlay = widgets.ABDropOverlay(self)
            self.overlay.show()
            event.accept()

    def dragLeaveEvent(self, event):
        """
        Handles the drag leave event.

        Args:
            event: The drag leave event.
        """
        # Reset the palette on drag leave
        self.overlay.deleteLater()

    def dropEvent(self, event):
        """
        Handles the drop event.

        Args:
            event: The drop event.
        """
        logger.debug(f"Dropevent from: {type(event.source()).__name__} to: {self.__class__.__name__}")
        # Reset the palette on drop
        self.overlay.deleteLater()

        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
        exchanges = {"technosphere": set(), "biosphere": set()}

        for key in keys:
            if exc_type := get_exchange_type(key):
                exchanges[exc_type].add(key)

        # Run the action for new exchanges
        for exc_type, keys in exchanges.items():
            app.actions.ExchangeNew.run(keys, self.parent().activity.key, exc_type)


class GraphBackend(QObject):
    """
    A backend object for communication between Python and JavaScript.
    This object is exposed to the JavaScript side and provides methods
    that can be called from JavaScript to control the graph.
    """
    def __init__(self, graph_tab: GraphTab, parent=None):
        """
        Initializes the GraphBackend object.

        Args:
            graph_tab (GraphTab): The GraphTab widget this backend is associated with.
            parent (QObject, optional): The parent object. Defaults to None.
        """
        super().__init__(parent)
        self.graph_tab = graph_tab

    @Slot(str)
    def expand_node(self, node_id: str):
        """
        Expands a node in the graph.

        Args:
            node_id (str): The ID of the node to expand.
        """
        self.graph_tab.expand_node(node_id)

    @Slot(str)
    def collapse_node(self, node_id: str):
        """
        Collapses a node in the graph.

        Args:
            node_id (str): The ID of the node to collapse.
        """
        self.graph_tab.collapse_node(node_id)


class Bridge(QObject):
    """
    A bridge for communication between Python and JavaScript.

    Attributes:
        update_graph (SignalInstance): A signal to update the graph.
        ready (SignalInstance): A signal indicating that the bridge is ready.
    """
    update_graph: SignalInstance = Signal(str)
    ready: SignalInstance = Signal()

    @Slot()
    def is_ready(self):
        """
        Emits the ready signal.
        """
        self.ready.emit()


class Page(QtWebEngineWidgets.QWebEnginePage):
    """
    A web engine page to display the HTML content.

    Methods:
        javaScriptConsoleMessage: Logs JavaScript console messages.
    """
    def javaScriptConsoleMessage(self, level: QtWebEngineWidgets.QWebEnginePage.JavaScriptConsoleMessageLevel, message: str, line: str, _: str):
        """
        Logs JavaScript console messages.

        Args:
            level (QtWebEngineWidgets.QWebEnginePage.JavaScriptConsoleMessageLevel): The message level.
            message (str): The message content.
            line (str): The line number.
            _ (str): Unused parameter.
        """
        if level == QtWebEngineWidgets.QWebEnginePage.InfoMessageLevel:
            logger.info(f"JS Info (Line {line}): {message}")
        elif level == QtWebEngineWidgets.QWebEnginePage.WarningMessageLevel:
            logger.warning(f"JS Warning (Line {line}): {message}")
        elif level == QtWebEngineWidgets.QWebEnginePage.ErrorMessageLevel:
            logger.error(f"JS Error (Line {line}): {message}")
        else:
            logger.debug(f"JS Log (Line {line}): {message}")



from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt
import pandas as pd

from activity_browser.ui import widgets, core, delegates, icons
from activity_browser.app import metadata
from activity_browser.bwutils.commontasks import refresh_node


class NodeSelectDialog(QtWidgets.QDialog):
    node_selected = QtCore.Signal(dict)

    def __init__(self, parent=None, drag_enabled=False):
        super().__init__(parent)

        self.setWindowFlags(
            QtCore.Qt.WindowType.Popup |
            QtCore.Qt.WindowType.FramelessWindowHint
        )
        self.setFixedWidth(400)
        self.setSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum)

        self.edit = widgets.ABLineEdit(self)
        self.edit.setPlaceholderText("Enter text to search for a node")
        self.edit.textChangedDebounce.connect(self.on_search)

        # Create model and tree view for results
        self.model = NodeSearchModel(parent=self)
        self.tree_view = NodeSearchView(self)
        self.tree_view.setModel(self.model)

        self.tree_view.clicked.connect(self.accept)
        self.tree_view.dragStarted.connect(self.on_drag_started)
        self.tree_view.setDragEnabled(drag_enabled)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(5, 5, 5, 0)
        layout.addWidget(self.edit)
        layout.addWidget(self.tree_view)
        self.setLayout(layout)

        self.setFixedHeight(self.sizeHint().height())

    def showEvent(self, event):
        super().showEvent(event)
        self.edit.setFocus()

    def on_search(self, text: str):
        if not text.strip():
            # Clear results
            self.model.set_dataframe(pd.DataFrame())
            self.tree_view.setFixedHeight(0)
            self.adjustSize()
            self.setFixedHeight(self.sizeHint().height())
            return

        # Search and get results
        result_df = metadata.search(text)
        result_df = result_df[0:10] if len(result_df) > 10 else result_df

        # Add a placeholder "node" column for the CardDelegate
        result_df["node"] = None

        # Update model with search results
        self.model.set_dataframe(result_df)

        # Adjust height based on results
        if len(result_df) > 0:
            self.tree_view.setFixedHeight(min(400, len(result_df) * 80 + 20))
        else:
            self.tree_view.setFixedHeight(0)

        # Adjust dialog to minimum size
        self.adjustSize()
        self.setFixedHeight(self.sizeHint().height())

    def on_drag_started(self):
        """Handle when a drag operation is started"""
        self.hide()  # Close the dialog

    def get_selected_node(self):
        """Return the currently selected node data"""
        index = self.tree_view.currentIndex()
        if not index.isValid():
            return None
        node_id = self.model.get(index, "id")
        if not node_id:
            return None
        return refresh_node(node_id)


class NodeSearchModel(core.ABTreeModel):
    """Model for displaying search results in the node select dialog."""

    def columns(self) -> list[str]:
        return ["index", "node"]

    def indexDragEnabled(self, index: QtCore.QModelIndex) -> bool:
        return True

    def displayData(self, index: QtCore.QModelIndex) -> any:
        if not index.isValid():
            return None

        column_name = self.columns()[index.column()]
        if not column_name == "node":
            return super().displayData(index)

        row_data = self.row(index).copy()
        row_data.dropna(inplace=True)

        # Get the product or name for title
        title = row_data.get("product") or row_data.get("name")

        # Build subtitle with type and database
        if row_data.get("categories"):
            subtitle = ", ".join([str(cat) for cat in row_data.get("categories")])
        elif row_data.get("product"):
            subtitle = row_data.get("name")
        else:
            subtitle = ""

        # Build categories list from unit, location
        categories = []
        if row_data.get("unit"):
            categories.append(str(row_data.get("unit")))
        if row_data.get("location"):
            categories.append(str(row_data.get("location")))
        if row_data.get("database"):
            categories.append(str(row_data.get("database")))

        return {
            "title": title,
            "subtitle": subtitle,
            "categories": categories if categories else None,
        }

    def decorationData(self, index: QtCore.QModelIndex) -> QtGui.QIcon:
        if not index.isValid():
            return icons.qicons.empty

        node_type = self.get(index, "type")

        if node_type == "product":
            return icons.qicons.product
        if node_type == "waste":
            return icons.qicons.waste
        if node_type == "processwithreferenceproduct":
            return icons.qicons.processproduct
        if node_type in ["natural resource", "emission", "inventory indicator", "economic", "social"]:
            return icons.qicons.biosphere
        return icons.qicons.process

    def mimeData(self, indices: list[QtCore.QModelIndex]):
        """
        Returns the mime data for the given indices.

        Args:
            indices (list[QtCore.QModelIndex]): The indices to get the mime data for.

        Returns:
            core.ABMimeData: The mime data.
        """
        data = core.ABMimeData()
        keys = [self.row(index).get("key") for index in indices if index.isValid()]
        keys = {key for key in keys if isinstance(key, tuple)}
        data.setPickleData("application/bw-nodekeylist", list(keys))
        return data


class NodeSearchView(widgets.ABTreeView):
    """Tree view for displaying node search results."""
    dragStarted: QtCore.SignalInstance = QtCore.Signal()

    defaultColumnDelegates = {
        "node": delegates.CardDelegate,
    }

    def __init__(self, parent: NodeSelectDialog):
        super().__init__(parent)
        self.setSelectionBehavior(widgets.ABTreeView.SelectionBehavior.SelectRows)
        self.setSelectionMode(widgets.ABTreeView.SelectionMode.SingleSelection)
        self.viewport().setBackgroundRole(QtGui.QPalette.ColorRole.Window)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.setHeaderHidden(True)

        self.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)
        self.setFixedHeight(0)


    def startDrag(self, supportedActions: Qt.DropAction) -> None:
        self.dragStarted.emit()
        super().startDrag(supportedActions)




import datetime
from loguru import logger

from qtpy import QtWidgets, QtGui, QtCore
from qtpy.QtCore import Qt

import bw2data as bd
import pandas as pd

from activity_browser import app
from activity_browser.bwutils.commontasks import count_database_records
from activity_browser.ui import widgets, icons, delegates, core
from activity_browser.app.menu_bar import ImportDatabaseMenu


class DatabasesPane(widgets.ABAbstractPane):
    """
    A widget that displays the databases and their details.

    Attributes:
        view (DatabasesView): The view displaying the databases.
        model (DatabasesModel): The model containing the data for the databases.
    """
    title = "Databases"
    unique = True

    def __init__(self, parent):
        """
        Initializes the DatabasesPane widget.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
        """
        super().__init__(parent)
        self._populate_later_flag = False

        self.model = DatabasesModel(parent=self)
        self.view = DatabasesView()
        self.view.setModel(self.model)

        self.view.setAlternatingRowColors(True)
        self.view.setIndentation(0)

        self.build_layout()
        self.connect_signals()

    def connect_signals(self):
        """
        Connects the signals to the appropriate slots.
        """
        app.signals.meta.databases_changed.connect(self.syncLater)
        app.signals.metadata.synced.connect(self.syncLater)
        app.signals.database.deleted.connect(self.syncLater)
        app.signals.database_read_only_changed.connect(self.syncLater)

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(5, 0, 5, 5)
        self.setLayout(layout)

    def syncLater(self):
        """
        Schedules a sync operation to be performed later.
        """

        def slot():
            self._populate_later_flag = False
            self.sync()
            self.thread().eventDispatcher().awake.disconnect(slot)

        if self._populate_later_flag:
            return

        self._populate_later_flag = True
        self.thread().eventDispatcher().awake.connect(slot)

    def sync(self):
        """
        Synchronizes the model with the current state of the databases.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        df = self.build_df()
        self.model.set_dataframe(df)
        self.view.resizeColumnToContents(1)
        self.view.header().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Fixed)

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the databases.

        Returns:
            pd.DataFrame: The DataFrame containing the databases data.
        """
        data = []
        for name in bd.databases:
            # get the modified time, in case it doesn't exist, just write 'now' in the correct format
            dt = bd.databases[name].get("modified", datetime.datetime.now().isoformat())
            dt = datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S.%f")

            # final column includes interactive checkbox which shows read-only state of db
            data.append(
                {
                    "name": name,
                    "depends": ", ".join(bd.databases[name].get("depends", [])),
                    "modified": dt,
                    "records": count_database_records(name),
                    "read_only": bd.databases[name].get("read_only", True),
                    "default_allocation": bd.databases[name].get("default_allocation", "unspecified"),
                    "backend": bd.databases[name].get("backend")
                }
            )

        cols = ["read_only", "name", "records", "depends", "default_allocation", "modified", "backend"]

        return pd.DataFrame(data, columns=cols)


class DatabasesView(widgets.ABTreeView):
    """
    A view that displays the databases in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "modified": delegates.DateTimeDelegate,
    }

    class ExportDatabaseContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m: m.setTitle("Export database" if len(m.parent().selected_databases) == 1 else "Export databases"),
            lambda m, p: m.add(app.actions.DatabaseExportExcel, p.selected_databases if p.selected_databases else [],
                               enable=len(p.selected_databases) >= 1,
                               text="to .xlsx",
                               ),
            lambda m, p: m.add(app.actions.DatabaseExportBW2Package, p.selected_databases if p.selected_databases else [],
                               enable=len(p.selected_databases) >= 1,
                               text="to .bw2package",
                               ),
        ]

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(app.actions.DatabaseNew),
            lambda m: m.addMenu(ImportDatabaseMenu(m)),
            lambda m, p: m.addMenu(DatabasesView.ExportDatabaseContextMenu(parent=p)),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(app.actions.DatabaseDelete, p.selected_databases if p.selected_databases else [],
                               enable=len(p.selected_databases) >= 1,
                               text="Delete databases" if len(p.selected_databases) > 1 else "Delete database",
                               ),
            lambda m, p: m.add(app.actions.DatabaseDuplicate, p.selected_databases[0] if p.selected_databases else None,
                               enable=len(p.selected_databases) == 1),
            lambda m, p: m.add(app.actions.DatabaseRelink, p.selected_databases[0] if p.selected_databases else None),
            lambda m, p: m.add(app.actions.DatabaseProcess, p.selected_databases[0] if p.selected_databases else None,
                               enable=len(p.selected_databases) == 1),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(app.actions.DatabaseSetReadonly, p.selected_databases[0] if p.selected_databases else None,
                               not m.selected_readonly,
                               enable=len(p.selected_databases) == 1,
                               text="Unlock database" if m.selected_readonly else "Lock database",
                               ),
        ]

        @property
        def selected_readonly(self):
            """
            Returns the read-only state of the selected database.

            Returns:
                bool: The read-only state of the selected database.
            """
            if not self.parent().selected_databases:
                return None
            index = self.parent().selectedIndexes()[0]
            row = self.parent().model().row(index)
            return row.get("read_only") if row is not None else None

    class HeaderMenu(QtWidgets.QMenu):
        """
        A header menu for the DatabasesView. Currently not used.
        """

        def __init__(self, *args, **kwargs):
            super().__init__()

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        """
        Handles the mouse double click event to toggle the read-only state or select the database.

        Args:
            event (QtGui.QMouseEvent): The mouse double click event.
        """
        index = self.indexAt(event.pos())

        if not index.isValid():
            return super().mouseDoubleClickEvent(event)

        row = self.model().row(index)
        if row is None:
            return super().mouseDoubleClickEvent(event)

        db_name = row.get("name")

        if index.column() == 1:
            read_only = row.get("read_only")
            app.actions.DatabaseSetReadonly.run(db_name, not read_only)
            return

        app.actions.DatabaseOpen.run([db_name])

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        """
        Handles key press events. Specifically handles the Delete key to delete selected databases.

        Args:
            event (QtGui.QKeyEvent): The key press event.
        """
        if event.key() == Qt.Key_Delete:
            if self.selected_databases:
                app.actions.DatabaseDelete.run(self.selected_databases)
                return
        
        super().keyPressEvent(event)

    @property
    def selected_databases(self) -> list:
        """
        Returns the database name of the user-selected index.

        Returns:
            str: The name of the selected database.
        """
        if not self.selectedIndexes():
            return []
        names = self.model().values_from_indices("name", self.selectedIndexes())
        return list(set(names))


class DatabasesModel(core.ABTreeModel):
    """
    A model representing the data for the databases.
    """

    def decorationData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides decoration data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide decoration data.

        Returns:
            The decoration data for the index.
        """
        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return None

        if column_name == "read_only":
            return icons.qicons.locked if row.get("read_only") else icons.qicons.empty

        return None

    def displayData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides display data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide display data.

        Returns:
            The display data for the index.
        """
        column_name = self.column_name(index)
        row = self.row(index)

        if row is None:
            return None

        if column_name == "read_only":
            return None

        return row.get(column_name)

    def fontData(self, index: QtCore.QModelIndex) -> any:
        """
        Provides font data for the model.

        Args:
            index (QtCore.QModelIndex): The index for which to provide font data.

        Returns:
            QtGui.QFont: The font data for the index.
        """
        column_name = self.column_name(index)

        if column_name == "name":
            font = QtGui.QFont()
            font.setWeight(QtGui.QFont.Weight.DemiBold)
            return font

        return None

    def headerData(self, section, orientation=Qt.Orientation.Horizontal, role=Qt.ItemDataRole.DisplayRole):
        """
        Provides header data for the model.

        Args:
            section (int): The section index.
            orientation (Qt.Orientation): The orientation of the header.
            role (Qt.ItemDataRole): The role for which to provide header data.

        Returns:
            The header data for the model.
        """
        if section == 1 and role == Qt.ItemDataRole.DisplayRole:
            return ""
        if section == 1 and role == Qt.ItemDataRole.DecorationRole:
            return icons.qicons.unlocked
        return super().headerData(section, orientation, role)



from functools import partial
from typing import List

from activity_browser import app
from activity_browser.app.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons
from activity_browser.ui.dialogs import UncertaintyDialog


class CFUncertaintyModify(ABAction):
    """
    ABAction to launch the UncertaintyDialog for Characterization Factor and handles the output by writing the
    uncertainty data using the ImpactCategoryController to the Characterization Factor in question.
    """

    icon = qicons.edit
    text = "Modify uncertainty"

    @classmethod
    @exception_dialogs
    def run(cls, method_name: tuple, char_factors: List[tuple], uncertainty_dict: dict = None):

        if uncertainty_dict is None:
            initial = char_factors[0][1]
            initial = initial if isinstance(initial, dict) else {}

            ok, uncertainty_dict = UncertaintyDialog.get_uncertainty_dict(
                parent=app.main_window,
                initial=initial,
                )

            if not ok:
                return
        
        method = bd.Method(method_name)
        method_dict = {cf[0]: cf[1] for cf in method.load()}

        for cf in char_factors:
            if isinstance(cf[1], dict):
                cf[1].update(uncertainty_dict)
                method_dict[cf[0]] = cf[1]
            else:
                uncertainty_dict["amount"] = cf[1]
                method_dict[cf[0]] = uncertainty_dict
        
        method.write(list(method_dict.items()))



from qtpy import QtWidgets

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure


class ABPlot(QtWidgets.QWidget):
    ALL_FILTER = "All Files (*.*)"
    PNG_FILTER = "PNG (*.png)"
    SVG_FILTER = "SVG (*.svg)"

    def __init__(self, parent=None):
        super().__init__(parent)
        # create figure, canvas, and axis
        self.figure = Figure(constrained_layout=True)
        self.canvas = FigureCanvasQTAgg(self.figure)
        self.canvas.setMinimumHeight(0)

        self.ax = self.figure.add_subplot(111)  # create an axis
        self.plot_name = "Figure"

        # set the layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.canvas)
        self.setLayout(layout)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.updateGeometry()

    def plot(self, *args, **kwargs):
        raise NotImplementedError

    def reset_plot(self) -> None:
        self.figure.clf()
        self.ax = self.figure.add_subplot(111)

    def get_canvas_size_in_inches(self):
        return tuple(x / self.figure.dpi for x in self.canvas.get_width_height())

    def to_png(self):
        """Export to .png format."""
        from activity_browser.bwutils.commontasks import savefilepath

        filepath = savefilepath(
            default_file_name=self.plot_name, file_filter=self.PNG_FILTER
        )
        if filepath:
            if not filepath.endswith(".png"):
                filepath += ".png"
            self.figure.savefig(filepath)

    def to_svg(self):
        """Export to .svg format."""      
        from activity_browser.bwutils.commontasks import savefilepath
        
        filepath = savefilepath(
            default_file_name=self.plot_name, file_filter=self.SVG_FILTER
        )
        if filepath:
            if not filepath.endswith(".svg"):
                filepath += ".svg"
            self.figure.savefig(filepath)




# -*- coding: utf-8 -*-
"""
bwutils is a collection of methods that build upon brightway2 and are generic enough to provide here so that we avoid
re-typing the same code in different parts of the Activity Browser.
"""




from loguru import logger

import bw2data as bd
from bw2data.parameters import (ActivityParameter, Group,
                                                     GroupDependency,
                                                     parameters)

from activity_browser import app
from activity_browser.app.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ParameterGroupDelete(ABAction):
    """
    ABAction to delete an existing parameter.
    """

    icon = qicons.delete
    text = "Delete parameter group..."

    @staticmethod
    @exception_dialogs
    def run(parameter_groups: list[str]):
        for group in parameter_groups:
            if group in ["project"] + list(bd.databases):
                logger.warning(f"Cannot delete built-in parameter group '{group}'. Skipping.")
                continue

            group_entry = Group.get(Group.name == group)

            # Delete all parameters in the group
            params_in_group = ActivityParameter.select().where(ActivityParameter.group == group)
            if any([ActivityParameter.is_dependent_on(p.name, p.group) for p in params_in_group]):
                raise Exception(f"Cannot delete parameter group '{group}' because some parameters are dependencies for other parameters.")

            for param in params_in_group:
                param.delete_instance()

            # Delete group dependencies
            GroupDependency.delete().where(GroupDependency.group == group).execute()
            # Delete the group itself
            group_entry.delete_instance()


        # After deleting things, recalculate and signal changes
        parameters.recalculate()

        # No fire when everything is still fresh after recalculation, so need to fire manually to be sure everything is
        # updated correctly.
        app.signals.parameter.recalculated.emit()



from loguru import logger

from qtpy.QtWebEngineWidgets import QWebEnginePage


class ABWebEnginePage(QWebEnginePage):
    def javaScriptConsoleMessage(self, level: QWebEnginePage.JavaScriptConsoleMessageLevel, message: str, line: str, _: str):
        if level == QWebEnginePage.InfoMessageLevel:
            logger.info(f"JS Info (Line {line}): {message}")
        elif level == QWebEnginePage.WarningMessageLevel:
            logger.warning(f"JS Warning (Line {line}): {message}")
        elif level == QWebEnginePage.ErrorMessageLevel:
            logger.error(f"JS Error (Line {line}): {message}")
        else:
            logger.debug(f"JS Log (Line {line}): {message}")



import sys

# This patch fixes a bug on Python 3.10 where shiboken6 patches the typing module with a broken Self type. Only way to
# fix it is to patch it ourselves first.
if sys.version_info[1] < 11:
    import typing
    if hasattr(typing, "Self"): # if Self has already been patched into typing, delete it first
        del typing.Self
    import typing_extensions
    setattr(typing, "Self", typing_extensions.Self)

try:
    import PySide6
except ImportError:
    import qtpy

def setup_logging():
    """Configure loguru sinks for console and file logging."""
    from loguru import logger
    import os
    import platformdirs

    logger.level("SYNC", no=9, color="<cyan>")
    logger.level("SIGNAL", no=19, color="<yellow>")
    logger.level("TEST", no=19, color="<cyan>")

    logger.remove()
    logger.add(sys.stderr, level=6, colorize=True,
               format="<green>{time:HH:mm:ss}</green> | <level>{level: <8}</level> | <level>{message}</level>")

    log_dir = platformdirs.user_log_dir(appname="ActivityBrowser", appauthor="pylca")
    os.makedirs(log_dir, exist_ok=True)
    log_file = os.path.join(log_dir, "activity_browser.log")
    logger.add(log_file, level="DEBUG", rotation="5 MB", retention=5)

def run_activity_browser():
    from .__main__ import run_activity_browser

setup_logging()


# -*- coding: utf-8 -*-
from .settings_page import SettingsPage
from .base import BaseSettingsChapter
from .startup import StartupSettingsChapter
from .appearance import AppearanceSettingsChapter
from .project_manager import ProjectManagerSettingsChapter
from .metadatastore import MetadataStoreSettingsChapter
from .plugins import PluginsSettingsChapter

__all__ = [
    "SettingsPage", 
    "BaseSettingsChapter",
    "StartupSettingsChapter",
    "AppearanceSettingsChapter",
    "ProjectManagerSettingsChapter",
    "MetadataStoreSettingsChapter",
    "PluginsSettingsChapter",
]



from qtpy import QtWidgets
from loguru import logger

import bw2data as bd
import pandas as pd

from activity_browser import app
from activity_browser.ui import widgets, delegates, core


class ImpactCategorySection(QtWidgets.QWidget):
    def __init__(self, calculation_setup_name: str, parent=None):
        super().__init__(parent)

        self.calculation_setup_name = calculation_setup_name
        self.calculation_setup = bd.calculation_setups.get(self.calculation_setup_name)

        self.view = ImpactCategoryView(self)
        self.model = ImpactCategoryModel(parent=self)
        self.view.setModel(self.model)

        self.build_layout()

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        self.setLayout(layout)

    def sync(self):
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        try:
            self.calculation_setup = bd.calculation_setups[self.calculation_setup_name]
            df = self.build_df()
            df.reset_index(drop=True, inplace=True)
            self.model.set_dataframe(df)
        except KeyError:
            self.parent().close()
            self.parent().deleteLater()

    def build_df(self):
        data = [bd.methods.get(method_name) for method_name in self.calculation_setup.get("ia", [])]
        df = pd.DataFrame(data, columns=["name", "unit", "num_cfs"])

        df["name"] = self.calculation_setup.get("ia", [])
        df["_cs_name"] = self.calculation_setup_name

        cols = ["name", "unit", "num_cfs", "_cs_name"]

        return df[cols]


class ImpactCategoryView(widgets.ABTreeView):
    defaultColumnDelegates = {
        "name": delegates.StringDelegate
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(app.actions.CSDeleteImpactCategory, m.cs_name, m.selected_ics,
                               text="Delete Impact Category" if len(m.selected_ics) == 1 else "Delete Impact Categories",
                               enable=len(m.selected_ics) > 0
                               ),
        ]

        @property
        def selected_ics(self):
            return list(set([index.row() for index in self.parent().selectedIndexes()]))

        @property
        def cs_name(self):
            return self.parent().parent().calculation_setup_name

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

    def dragMoveEvent(self, event) -> None:
        pass

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("application/bw-methodnamelist"):
            event.accept()

    def dropEvent(self, event) -> None:
        event.accept()
        cs_name = self.parent().calculation_setup_name
        method_names = event.mimeData().retrievePickleData("application/bw-methodnamelist")
        app.actions.CSAddImpactCategory.run(cs_name, method_names)


class ImpactCategoryModel(core.ABTreeModel):
    """
    A model representing the data for the impact categories.
    """
    pass
